import types

from django.db.models import Model, QuerySet, Field

from chamber.utils import get_class_method, call_function_with_unknown_input, InvalidFunctionArguments

from pyston.serializer import get_resource_class_or_none

from is_core.forms.utils import ReadonlyValue
from is_core.forms.widgets import ReadonlyWidget, ManyToManyReadonlyWidget, ModelObjectReadonlyWidget
from is_core.utils import LOOKUP_SEP, get_field_from_model_or_none


class GetFieldDescriptorException(Exception):
    pass


class FieldDoesNotExist(GetFieldDescriptorException):
    pass


class FieldIsNotRelation(GetFieldDescriptorException):
    pass


class GetFieldDescriptorValueError(GetFieldDescriptorException):
    pass


def pretty_name(value):
    return value.replace('_', ' ')


registered_model_descriptors = {}


def register_model_descriptor(model_class, descriptor_class):
    registered_model_descriptors[model_class] = descriptor_class


def get_model_descriptor(model_class):
    for base_model_class, descriptor_class in registered_model_descriptors.items():
        if issubclass(model_class, base_model_class):
            return descriptor_class
    return None


class FieldDescriptor:

    is_method = False
    is_model_field = False
    is_property = False

    @classmethod
    def init_descriptor_or_none(cls, model, field_name, view):
        raise NotImplementedError

    def __init__(self, model, field_name, model_field_or_method):
        self.model = model
        self.field_name = field_name
        self.model_field_or_method = model_field_or_method

    def get_related_model(self):
        related_model = getattr(self.model_field_or_method, 'related_model', None)
        if related_model:
            return related_model
        else:
            raise FieldIsNotRelation('Model ("{}") field ("{}") with name "{}" is not relation'.format(
                self.model, self.model_field_or_method, self.field_name
            ))

    def get_label(self):
        raise NotImplementedError

    def get_widget(self):
        return ReadonlyWidget

    def get_value(self, instance, request=None):
        raise NotImplementedError

    def _get_verbose_value(self, raw, instance):
        if hasattr(self.model_field_or_method, 'humanized') and self.model_field_or_method.humanized:
            return self.model_field_or_method.humanized(raw, instance)
        elif (hasattr(self.model_field_or_method, 'attname')
                and hasattr(instance, 'get_{}_display'.format(self.model_field_or_method.attname))):
            return getattr(instance, 'get_{}_display'.format(self.model_field_or_method.attname))()
        else:
            return None

    def get_readonly_value(self, instance, request=None):
        value = self.get_value(instance, request)
        verbose_value = self._get_verbose_value(value, instance)
        return ReadonlyValue(value, verbose_value) if verbose_value is not None else value


class ModelFieldDescriptorBase(type):

    def __new__(cls, name, bases, attrs):
        new_class = type.__new__(cls, name, bases, attrs)
        if new_class.base_model:
            register_model_descriptor(new_class.base_model, new_class)
        return new_class


class BaseModelFieldDescriptor(FieldDescriptor, metaclass=ModelFieldDescriptorBase):

    is_model_field = True
    base_model = None


class DjangoFieldDescriptor(BaseModelFieldDescriptor):
    """
    Field descriptor which gets field data from Django model field.
    """

    is_model_field = True
    base_model = Model

    @classmethod
    def init_descriptor_or_none(cls, model, field_name, view):
        field = get_field_from_model_or_none(model, field_name)
        return cls(model, field_name, field) if field else None

    def get_label(self):
        if self.model_field_or_method.auto_created and self.model_field_or_method.is_relation:
            return self.model_field_or_method.field.reverse_verbose_name
        elif hasattr(self.model_field_or_method, 'verbose_name'):
            return self.model_field_or_method.verbose_name
        else:
            return pretty_name(self.field_name)

    def get_widget(self):
        field = self.model_field_or_method

        if field.auto_created and (field.one_to_many or field.many_to_many):
            return ManyToManyReadonlyWidget
        elif field.auto_created and field.one_to_one:
            return ModelObjectReadonlyWidget
        elif isinstance(field, Field):
            form_field = field.formfield()
            return form_field.readonly_widget if form_field else ReadonlyWidget
        else:
            # GenericForeignKey is model field without formfield method
            return ReadonlyWidget

    def get_value(self, instance, request=None):
        field = self.model_field_or_method
        if (field.auto_created and field.one_to_many) or field.many_to_many:
            return [obj for obj in getattr(instance, field.name).all()]
        elif field.auto_created and field.one_to_one:
            return None if not hasattr(instance, field.name) else getattr(instance, field.name)
        else:
            return getattr(instance, field.name) if hasattr(instance, field.name) else None


class MethodOrPropertyDescriptor(FieldDescriptor):
    """
    Field descriptor which gets field data from object method or property.
    """

    def get_label(self):
        return getattr(self.model_field_or_method, 'short_description', pretty_name(self.field_name))

    def _get_field_or_method_value(self, instance, request=None):
        raise NotImplementedError

    def _clean_value(self, value):
        related_model = getattr(self.model_field_or_method, 'related_model', None)
        if value is not None and related_model:
            if isinstance(value, (QuerySet, list, tuple, set, types.GeneratorType)):
                if any(not isinstance(v, related_model) for v in value):
                    raise GetFieldDescriptorValueError(
                        'Method "{}" not returned related model ({}) value or its iterable'.format(
                            self.model_field_or_method, related_model
                        )
                    )
                return list(value)
            elif not isinstance(value, related_model):
                raise GetFieldDescriptorValueError(
                    'Method "{}" not returned related model ({}) value or its iterable'.format(
                        self.model_field_or_method, related_model
                    )
                )
        return value

    def get_value(self, instance, request=None):
        return self._clean_value(self._get_field_or_method_value(instance, request))


class PropertyDescriptor(MethodOrPropertyDescriptor):
    """
    Field descriptor which gets field data from object property.
    """

    is_property = True

    @classmethod
    def init_descriptor_or_none(cls, model, field_name, view):
        property = getattr(model, field_name, None)
        return cls(model, field_name, property) if property else None

    def _get_field_or_method_value(self, instance, request=None):
        return getattr(instance, self.field_name)


class BaseModelMethodDescriptor(MethodOrPropertyDescriptor):
    """
    Field descriptor which gets field data from object method.
    """

    is_method = True

    def _get_field_or_method_value(self, instance, request=None):
        return call_function_with_unknown_input(
            self.model_field_or_method, **self._get_method_kwargs(instance, request)
        )

    def _get_method_kwargs(self, instance, request=None):
        if request:
            return {'obj': instance, 'request': request}
        else:
            return {'obj': instance}


class ModelMethodDescriptor(BaseModelMethodDescriptor):
    """
    Field descriptor which gets field data from Django model method.
    """
    @classmethod
    def init_descriptor_or_none(cls, model, field_name, view):
        method = get_class_method(model, field_name)
        return cls(model, field_name, method) if method else None

    def _get_method_kwargs(self, instance, request=None):
        return {
            'self': instance,
            **super()._get_method_kwargs(instance, request)
        }


class ResourceMethodDescriptor(BaseModelMethodDescriptor):
    """
    Field descriptor which gets field data from Django model resource method.
    """

    @classmethod
    def init_descriptor_or_none(cls, model, field_name, view):
        resource = get_resource_class_or_none(model)
        if not resource:
            return None
        resource_method = resource.get_method_returning_field_value(field_name)
        return cls(model, field_name, resource_method, resource) if resource_method else None

    def __init__(self, model, field_name, method, resource_class):
        super().__init__(model, field_name, method)
        self.resource_class = resource_class

    def _get_method_kwargs(self, instance, request=None):
        if request is None:
            raise InvalidFunctionArguments('For resource method "{}" value is required request'.format(self.field_name))

        return {
            'self': self.resource_class(request),
            **super()._get_method_kwargs(instance, request)
        }


class ViewMethodDescriptor(BaseModelMethodDescriptor):
    """
    Field descriptor which gets field data from core view method.
    """

    @classmethod
    def init_descriptor_or_none(cls, model, field_name, view):
        if not view:
            return None
        view_method = view.get_method_returning_field_value(field_name)
        return cls(model, field_name, view_method, view) if view_method else None

    def __init__(self, model, field_name, method, view):
        super().__init__(model, field_name, method)
        self.view = view

    def _get_method_kwargs(self, instance, request=None):
        return {
            'self': self.view,
            **super()._get_method_kwargs(instance, request)
        }


def _get_field_descriptor(model, field_name, view=None):
    descriptor_classes = [ViewMethodDescriptor, ResourceMethodDescriptor, ModelMethodDescriptor]
    model_field_descriptor = get_model_descriptor(model)
    if model_field_descriptor:
        descriptor_classes.append(model_field_descriptor)
    descriptor_classes += [PropertyDescriptor]

    for descriptor_class in descriptor_classes:
        descriptor = descriptor_class.init_descriptor_or_none(model, field_name, view)
        if descriptor:
            return descriptor

    raise FieldDoesNotExist('Model ("{}") field with name "{}" was not found'.format(model, field_name))


def get_field_descriptors_from_path(model, field_path, view=None):
    """
    Helper returns list of field descriptors. Input field_path is consist of field names separated with "__".
    For every field name is found the right descriptor which describes how to get field label, value or
    how to humanize this value. There are 5 types of field descriptors:

    * ModelFieldDescriptor - value and label is get from django model field
    * PropertyDescriptor - value is get from instance property
    * ModelMethodDescriptor - value will be obtained as a result of model method call.
    * ResourceMethodDescriptor - firstly is found pyston resource of model. The result value is get from resource
                                 method.  Because resource is related with core. The method will be searching
                                 in the core too.
    * ViewMethodDescriptor - if view is in the input parameters. Descriptor can be get from view method too.

    :param model: Django model class.
    :param field_path: field names separated with __.
    :param view: view instance.
    :return: list of FieldDescriptor or GetFieldDescriptorException is raised
    """
    try:
        if LOOKUP_SEP in field_path:
            current_field_name, next_field_path = field_path.split(LOOKUP_SEP, 1)
            field_descriptor = _get_field_descriptor(
                model, current_field_name, view
            )
            return [field_descriptor] + get_field_descriptors_from_path(
                field_descriptor.get_related_model(), next_field_path
            )
        else:
            return [_get_field_descriptor(model, field_path, view)]
    except (GetFieldDescriptorException, InvalidFunctionArguments) as ex:
        raise GetFieldDescriptorException('Field path "{}" cannot be get from model "{}". Reason: {}'.format(
            field_path,
            model,
            ex
        ))


def get_field_value_from_path(instance, field_path, request=None, view=None, return_readonly_value=False):
    """
    Return value from model instance defined by field_path. Path can contain relations defined via "__" separator.
    :param instance: django model instance.
    :param field_path: field path where field can be relations separated with "__"-
    :param request: django HTTP request
    :param view: django view which can contains method with field value.
    :param return_readonly_value: value can be automatically humanized if value is set to True.
    :return: model field path value or GetFieldDescriptorException
    """
    field_descriptors = get_field_descriptors_from_path(
        instance.__class__, field_path, view
    )
    field_descriptor = field_descriptors[0]
    try:
        if len(field_descriptors) > 1:
            value = field_descriptor.get_value(instance, request)
            if value is None:
                return value
            elif isinstance(value, list):
                return [
                    get_field_value_from_path(
                        v, field_path.split(LOOKUP_SEP, 1)[1], request, return_readonly_value=return_readonly_value
                    )
                    for v in value
                ]
            else:
                return get_field_value_from_path(
                    value, field_path.split(LOOKUP_SEP, 1)[1], request, return_readonly_value=return_readonly_value
                )
        else:
            if return_readonly_value:
                return field_descriptor.get_readonly_value(instance, request)
            else:
                return field_descriptor.get_value(instance, request)
    except (GetFieldDescriptorValueError, InvalidFunctionArguments) as ex:
        raise GetFieldDescriptorValueError('Value cannot be get from field path "{}" of model "{}". Reason: {}'.format(
            field_path,
            instance.__class__,
            ex
        ))



from copy import deepcopy

from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.http.response import HttpResponseRedirect, Http404
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _
from django.views.generic.edit import FormView
from django.contrib.messages.api import get_messages, add_message
from django.contrib.messages import constants

from chamber.utils.forms import formset_has_file_field
from chamber.utils import transaction

from is_core.auth.permissions import PermissionsSet, CoreUpdateAllowed, CoreAllowed, DEFAULT_PERMISSION
from is_core.auth.views import FieldPermissionViewMixin
from is_core.generic_views.base import DefaultModelCoreViewMixin
from is_core.utils import (
    flatten_fieldsets, get_readonly_field_data, get_inline_views_from_fieldsets, get_inline_views_opts_from_fieldsets,
    GetMethodFieldMixin, get_model_name, get_fieldsets_without_disallowed_fields
)
from is_core.generic_views.mixins import ListParentMixin
from is_core.generic_views.inlines.inline_form_views import InlineFormView
from is_core.response import JsonHttpResponse
from is_core.forms.models import smartmodelform_factory
from is_core.forms.fields import SmartReadonlyField
from is_core.forms import SmartModelForm


class BaseFormView(GetMethodFieldMixin, DefaultModelCoreViewMixin, FormView):

    view_type = 'default'
    fieldsets = None
    fields = None
    form_template = 'is_core/forms/default_form.html'
    template_name = 'is_core/generic_views/default_form.html'
    messages = {
        'success': _('Object was saved successfully.'),
        'error': _('Please correct the error below.'),
    }
    readonly_fields = None

    save_button_label = _('Save')
    save_button_title = ''
    save_button_disable_on_submit = True

    cancel_button_label = _('Back')
    cancel_button_title = ''
    cancel_button_disable_on_submit = True

    atomic_save_form = True

    def get_success_url(self, obj):
        """
        URL string for redirect after saving
        """

        return self.request.get_full_path()

    def get_has_file_field(self, form, **kwargs):
        return formset_has_file_field(form)

    def is_readonly(self):
        return not self.has_permission('post')

    def get_form(self, form_class=None):
        if form_class is None:
            form_class = self.get_form_class()
        form = form_class(**self.get_form_kwargs())
        form.readonly = self.is_readonly()

        for field_name, field in form.fields.items():
            self.form_field(form, field_name, field)
        return form

    def get_form_action(self):
        return self.request.get_full_path()

    def get_readonly_fields(self):
        return () if self.readonly_fields is None else self.readonly_fields

    def get_message_kwargs(self, obj):
        return {'obj': force_text(obj)}

    def get_message(self, msg_type_or_level, obj=None):
        msg_kwargs = {}
        if obj:
            msg_kwargs = self.get_message_kwargs(obj)

        msg_type = (isinstance(msg_type_or_level, int) and constants.DEFAULT_TAGS.get(msg_type_or_level) or
                    msg_type_or_level)
        return self.messages.get(msg_type) % msg_kwargs

    def is_changed(self, form, **kwargs):
        """Return true if form was changed"""
        return form.has_changed()

    def save_obj(self, obj, form, change):
        """
        Must be added for non model forms
        this method should save object or raise exception
        """

        raise NotImplementedError

    def save_form(self, form, **kwargs):
        """Contains formset save, prepare obj for saving"""

        obj = form.save(commit=False)
        change = obj.pk is not None
        self.save_obj(obj, form, change)
        if hasattr(form, 'save_m2m'):
            form.save_m2m()
        return obj

    @transaction.smart_atomic
    def _atomic_save_form(self, *args, **kwargs):
        return self.save_form(*args, **kwargs)

    def form_valid(self, form, msg=None, msg_level=None, **kwargs):
        try:
            if self.atomic_save_form:
                obj = self._atomic_save_form(form, **kwargs)
            else:
                obj = self.save_form(form, **kwargs)
            return self.success_render_to_response(obj, msg, msg_level)
        except ValidationError as ex:
            return self.form_invalid(form, msg=force_text(ex.message), **kwargs)

    def form_invalid(self, form, msg=None, msg_level=None, **kwargs):
        msg_level = msg_level or constants.ERROR
        msg = msg or self.get_message(msg_level)
        add_message(self.request, msg_level, msg)
        return self.render_to_response(self.get_context_data(form=form, msg=msg, msg_level=msg_level, **kwargs))

    def get_popup_obj(self, obj):
        return {'_obj_name': force_text(obj)}

    @property
    def is_ajax_form(self):
        """Return if form will be rendered for ajax"""
        return self.has_snippet()

    @property
    def is_popup_form(self):
        """Return if form will be rendnered as popup"""
        return 'popup' in self.request.GET

    @property
    def form_snippet_name(self):
        """Return name for name for snippet which surround form"""
        return '%s-%s' % (self.view_name, 'form')

    def get_form_class_names(self):
        class_names = ['-'.join((self.view_type, self.site_name,
                                 self.core.get_menu_group_pattern_name(), 'form',)).lower()]
        if self.is_ajax_form:
            class_names.append('ajax')
        return class_names

    def get_prefix(self):
        return '-'.join((self.view_type, self.site_name)).lower()

    def get_context_data(self, **kwargs):
        context_data = super().get_context_data(**kwargs)
        form = context_data['form']
        kwargs.pop('form', form)
        context_data.update({
            'view_type': self.view_type,
            'form_template': self.form_template,
            'form_class_names': self.get_form_class_names(),
            'has_file_field': self.get_has_file_field(form, **kwargs),
            'action': self.get_form_action(),
            'form_snippet_name': self.form_snippet_name,
            'buttons': self.get_buttons()
        })
        return context_data

    def get_buttons(self):
        buttons = {}
        for key, value in self.get_buttons_dict().items():
            buttons[key] = {
                'label': value.get('label'),
                'title': value.get('title'),
                'disable_on_submit': value.get('disable_on_submit'),
            }
        return buttons

    def get_buttons_dict(self):
        return {
            'save': {
                'label': self.save_button_label,
                'title': self.save_button_title,
                'disable_on_submit': self.save_button_disable_on_submit,
            },
            'cancel': {
                'label': self.cancel_button_label,
                'title': self.cancel_button_title,
                'disable_on_submit': self.cancel_button_disable_on_submit,
            }
        }

    def get_fields(self):
        fieldsets = self.get_fieldsets()
        return self.fields if fieldsets is None else flatten_fieldsets(fieldsets)

    def get_fieldsets(self):
        return self.fieldsets

    def generate_fieldsets(self, form):
        fieldsets = self.get_fieldsets()
        return [
            (None, {'fields': self.get_fields() or list(form.base_fields.keys())})
        ] if fieldsets is None else fieldsets

    def get_initial(self):
        initial = super().get_initial()
        initial['_user'] = self.request.user
        initial['_request'] = self.request
        return initial

    def form_field(self, form, field_name, form_field):
        return form_field

    def get(self, request, *args, **kwargs):
        form = self.get_form()
        fieldsets = self.generate_fieldsets(form)
        return self.render_to_response(self.get_context_data(form=form, fieldsets=fieldsets))

    def post(self, request, *args, **kwargs):
        form = self.get_form()
        fieldsets = self.generate_fieldsets(form)

        is_valid = form.is_valid()
        is_changed = self.is_changed(form)

        if is_valid and is_changed:
            return self.form_valid(form)
        else:
            if is_valid and not is_changed:
                return self.form_invalid(
                    form, msg=_('No changes have been submitted.'), msg_level=constants.INFO,
                    fieldsets=fieldsets
                )
            return self.form_invalid(form, fieldsets=fieldsets)

    def get_snippet_names(self):
        if self.is_popup_form:
            return [self.form_snippet_name]
        else:
            return super().get_snippet_names()

    def success_render_to_response(self, obj, msg, msg_level):
        msg_level = msg_level or constants.SUCCESS
        msg = msg or self.get_message(msg_level, obj)
        if self.is_popup_form:
            return JsonHttpResponse({'messages': {msg_level: msg}, 'obj': self.get_popup_obj(obj)})
        elif self.is_ajax_form:
            add_message(self.request, msg_level, msg)
            location = self.get_success_url(obj)
            response = JsonHttpResponse({'location': location}, status=202)
            response['Location'] = location
            return response
        else:
            add_message(self.request, msg_level, msg)
            return HttpResponseRedirect(self.get_success_url(obj))

    def render_to_response(self, context, **response_kwargs):
        if self.has_snippet():
            extra_content = response_kwargs['extra_content'] = response_kwargs.get('extra_content', {})
            extra_content_messages = {}
            for message in get_messages(self.request):
                extra_content_messages[message.tags] = force_text(message)
            if extra_content_messages:
                extra_content['messages'] = extra_content_messages
        return super().render_to_response(context, **response_kwargs)


class DjangoBaseFormView(FieldPermissionViewMixin, BaseFormView):

    model = None
    exclude = ()
    field_labels = None
    inline_views = None
    form_template = 'is_core/forms/model_default_form.html'
    show_buttons = True

    def _get_field_labels(self):
        return self.field_labels

    def pre_save_obj(self, obj, form, change):
        pass

    def post_save_obj(self, obj, form, change):
        pass

    def form_field(self, form, field_name, form_field):
        form_field = super().form_field(form, field_name, form_field)
        placeholder = self.model._ui_meta.placeholders.get(field_name, None)
        if placeholder:
            form_field.widget.placeholder = placeholder
        return form_field

    def get_message_kwargs(self, obj):
        return {'obj': force_text(obj), 'name': force_text(obj._meta.verbose_name)}

    def get_exclude(self):
        return self.exclude

    def generate_readonly_fields(self):
        return list(self.get_readonly_fields()) + list(self._get_readonly_fields_from_permissions())

    def get_inline_views(self):
        return self.inline_views

    def _filter_inline_form_views(self, inline_views):
        return [view for view in inline_views if isinstance(view, InlineFormView)]

    def generate_fieldsets(self, form):
        fieldsets = deepcopy(self.get_fieldsets())

        if fieldsets and self.get_inline_views():
            raise ImproperlyConfigured('You can define either inline views or fieldsets.')

        if fieldsets is None:
            fieldsets = [
                (None, {
                    'fields': self.get_fields() or form.base_fields.keys()
                })
            ]
            for inline_view in self.get_inline_views() or ():
                inline_view_inst = (
                    inline_view(self.request, self, form.instance) if isinstance(inline_view, type) else inline_view
                )
                if inline_view_inst.can_render():
                    # Only inline view that can be rendered is added to formset
                    fieldsets.append((
                        inline_view_inst.get_title(), {
                            'inline_view': inline_view,
                            'inline_view_inst': inline_view_inst
                        }
                    ))
        else:
            inline_view_opts = get_inline_views_opts_from_fieldsets(fieldsets)
            for inline_view_opt in inline_view_opts:
                inline_view = inline_view_opt['inline_view']
                inline_view_inst = (
                    inline_view(self.request, self, form.instance) if isinstance(inline_view, type) else inline_view
                )
                if inline_view_inst.can_render():
                    # Only inline view that can be rendered is added to formset
                    inline_view_opt['inline_view_inst'] = inline_view_inst

        return get_fieldsets_without_disallowed_fields(
            self.request, fieldsets, self._get_disallowed_fields_from_permissions() | set(self.get_exclude())
        )

    def get_form_class(self):
        fields = self.get_fields()
        readonly_fields = self.generate_readonly_fields()
        return self.generate_form_class(fields, readonly_fields)

    def get_form_class_base(self):
        return self.form_class or SmartModelForm

    def get_is_bulk(self):
        return False

    def generate_form_class(self, fields=None, readonly_fields=()):
        form_class = self.get_form_class_base()
        exclude = list(self.get_exclude())
        if hasattr(form_class, '_meta') and form_class._meta.exclude:
            exclude.extend(form_class._meta.exclude)
        return smartmodelform_factory(self.model, self.request, form=form_class, exclude=exclude, fields=fields,
                                      formfield_callback=self.formfield_for_dbfield,
                                      readonly_fields=readonly_fields,
                                      formreadonlyfield_callback=self.formfield_for_readonlyfield,
                                      readonly=self.is_readonly(),
                                      labels=self._get_field_labels(), is_bulk=self.get_is_bulk())

    def update_form_initial(self, form):
        # Only new instance can get data from request queryset
        if not form.instance.pk:
            form.initial.update({k: self.request.GET.get(k) for k in self.request.GET.keys()
                                 if k not in form.readonly_fields})
        return form

    def get_form(self, form_class=None):
        return self.update_form_initial(super().get_form(form_class=form_class))

    def formfield_for_dbfield(self, db_field, **kwargs):
        return db_field.formfield(**kwargs)

    def formfield_for_readonlyfield(self, name, **kwargs):
        def _get_readonly_field_data(instance):
            return get_readonly_field_data(
                instance, name, self.request, view=self, field_labels=self._get_field_labels()
            )
        return SmartReadonlyField(_get_readonly_field_data)

    def get_has_file_field(self, form, inline_form_views=None, **kwargs):
        if super().get_has_file_field(form, **kwargs):
            return True

        for inline_form_view in () if inline_form_views is None else inline_form_views:
            if inline_form_view.get_has_file_field():
                return True

        return False

    def get_show_buttons(self):
        return self.show_buttons

    def get_cancel_url(self):
        return None

    def has_save_button(self):
        return not self.is_readonly()

    def is_changed(self, form, inline_form_views, **kwargs):
        for inline_form_view_instance in inline_form_views:
            if inline_form_view_instance.has_changed():
                return True
        return form.has_changed()

    def get(self, request, *args, **kwargs):
        form = self.get_form()
        fieldsets = self.generate_fieldsets(form)
        inline_views = get_inline_views_from_fieldsets(fieldsets)
        inline_form_views = self._filter_inline_form_views(inline_views)
        return self.render_to_response(self.get_context_data(form=form, inline_views=inline_views,
                                                             fieldsets=fieldsets,
                                                             inline_form_views=inline_form_views))

    def post(self, request, *args, **kwargs):
        form = self.get_form()
        fieldsets = self.generate_fieldsets(form)
        inline_views = get_inline_views_from_fieldsets(fieldsets)
        inline_form_views = self._filter_inline_form_views(inline_views)

        inline_forms_is_valid = True
        for inline_form_view in inline_form_views:
            inline_forms_is_valid = inline_form_view.is_valid() and inline_forms_is_valid

        is_valid = form.is_valid()
        is_changed = self.is_changed(form, inline_form_views=inline_form_views)
        if is_valid and inline_forms_is_valid and is_changed:
            return self.form_valid(form, inline_form_views=inline_form_views,
                                   inline_views=inline_views, fieldsets=fieldsets)
        else:
            if is_valid and not is_changed:
                return self.form_invalid(
                    form,
                    inline_form_views=inline_form_views,
                    inline_views=inline_views,
                    msg=_('No changes have been submitted.'),
                    msg_level=constants.INFO,
                    fieldsets=fieldsets
                )
            return self.form_invalid(
                form,
                inline_form_views=inline_form_views,
                inline_views=inline_views,
                fieldsets=fieldsets
            )

    def get_obj(self, cached=True):
        return None

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['instance'] = self.get_obj(False)
        return kwargs

    def save_form(self, form, inline_form_views=None, **kwargs):
        inline_form_views = () if inline_form_views is None else inline_form_views
        pre_save_inline_form_views = [
            inline_form_view for inline_form_view in inline_form_views if inline_form_view.save_before_parent
        ]
        post_save_inline_form_views = [
            inline_form_view for inline_form_view in inline_form_views if not inline_form_view.save_before_parent
        ]

        obj = form.save(commit=False)
        change = obj.pk is not None

        self.pre_save_obj(obj, form, change)

        for inline_form_view in pre_save_inline_form_views:
            inline_form_view.form_valid(self.request)

        self.save_obj(obj, form, change)
        if hasattr(form, 'save_m2m'):
            form.save_m2m()

        for inline_form_view in post_save_inline_form_views:
            inline_form_view.form_valid(self.request)

        self.post_save_obj(obj, form, change)
        return obj

    def get_context_data(self, form=None, inline_form_views=None, **kwargs):
        context_data = super().get_context_data(form=form, inline_form_views=inline_form_views, **kwargs)

        show_buttons = self.get_show_buttons()
        context_data.update({
            'module_name': get_model_name(self.model),
            'cancel_url': self.get_cancel_url() if show_buttons else None,
            'show_save_button': show_buttons and self.has_save_button()
        })
        return context_data

    def get_popup_obj(self, obj):
        app_label = self.model._meta.app_label
        model_name = self.model._meta.object_name
        return {'_obj_name': force_text(obj), 'pk': obj.pk, '_model': '%s.%s' % (app_label, model_name)}


class DjangoCoreFormView(ListParentMixin, DjangoBaseFormView):

    show_save_and_continue = True

    save_button_title = _('Save and navigate to the list')

    save_and_continue_button_title = _('Save and stay on the same page')
    save_and_continue_button_label = _('Save and continue')
    save_and_continue_button_disable_on_submit = True

    cancel_button_title = _('Do not save and go back to the list')

    export_types = None

    def get_prefix(self):
        return '-'.join((self.view_type, self.site_name, self.core.get_menu_group_pattern_name())).lower()

    def get_buttons_dict(self):
        buttons_dict = super().get_buttons_dict()
        buttons_dict.update({
            'save_and_continue': {
                'label': self.save_and_continue_button_label,
                'title': self.save_and_continue_button_title,
                'disable_on_submit': self.save_and_continue_button_disable_on_submit,
            }
        })
        return buttons_dict

    def save_obj(self, obj, form, change):
        self.core.save_model(self.request, obj, form, change)

    def pre_save_obj(self, obj, form, change):
        self.core.pre_save_model(self.request, obj, form, change)

    def post_save_obj(self, obj, form, change):
        self.core.post_save_model(self.request, obj, form, change)

    def get_readonly_fields(self):
        if self.readonly_fields is not None:
            return self.readonly_fields

        readonly_fields = self.core.get_readonly_fields(self.request, self.get_obj(True))
        return () if readonly_fields is None else readonly_fields

    def get_fieldsets(self):
        if self.fieldsets is not None or self.fields is not None:
            return self.fieldsets

        return self.core.get_fieldsets(self.request, self.get_obj(True))

    def get_fields(self):
        fieldsets = self.get_fieldsets()
        if fieldsets is not None:
            return flatten_fieldsets(fieldsets)
        elif self.fields is not None:
            return self.fields
        else:
            return self.core.get_fields(self.request, self.get_obj(True))

    def get_inline_views(self):
        if self.inline_views is not None:
            return self.inline_views

        return self.core.get_inline_views(self.request, self.get_obj(True))

    def _get_field_labels(self):
        return self.field_labels if self.field_labels is not None else self.core.get_field_labels(self.request)

    def get_form_class_base(self):
        obj = self.get_obj(True)
        return (
            self.form_class or
            (
                self.core.get_form_edit_class(self.request, obj)
                if obj
                else self.core.get_form_add_class(self.request, obj)
            )
        )

    def get_cancel_url(self):
        if ('list' in self.core.ui_patterns and
                self.core.ui_patterns.get('list').has_permission('get', self.request) and not self.has_snippet()):
            return self.core.ui_patterns.get('list').get_url_string(self.request)
        return None

    def has_save_and_continue_button(self):
        return (
            'list' in self.core.ui_patterns and not self.has_snippet() and
            self.core.ui_patterns.get('list').has_permission('get', self.request) and
            self.show_save_and_continue and
            not self.is_readonly()
        )

    def has_save_button(self):
        return self.view_type in self.core.ui_patterns and not self.is_readonly()

    def get_success_url(self, obj):
        if ('list' in self.core.ui_patterns and
                self.core.ui_patterns.get('list').has_permission('get', self.request) and
                'save' in self.request.POST):
            return self.core.ui_patterns.get('list').get_url_string(self.request)
        elif ('detail' in self.core.ui_patterns and
                self.core.ui_patterns.get('detail').has_permission('get', self.request, obj=obj) and
                'save-and-continue' in self.request.POST):
            return self.core.ui_patterns.get('detail').get_url_string(self.request, view_kwargs={'pk': obj.pk})
        else:
            return self.request.get_full_path()

    def get_context_data(self, form=None, inline_form_views=None, **kwargs):
        context_data = super().get_context_data(form=form, inline_form_views=inline_form_views, **kwargs)
        context_data.update({
            'show_save_and_continue': self.has_save_and_continue_button()
        })
        return context_data


class BulkChangeFormView(DjangoBaseFormView):

    form_template = 'is_core/views/bulk-change-view.html'
    is_ajax_form = False

    permission = PermissionsSet(
        get=CoreUpdateAllowed(),
        post=CoreUpdateAllowed(),
        **{
            DEFAULT_PERMISSION: CoreAllowed(),
        }
    )

    def get_prefix(self):
        return '-'.join((self.view_type, self.site_name, self.core.get_menu_group_pattern_name())).lower()

    def get_form_class_base(self):
        return self.form_class or self.core.get_rest_form_edit_class(self.request)

    def dispatch(self, request, *args, **kwargs):
        if 'snippet' not in request.GET:
            raise Http404
        return super().dispatch(request, *args, **kwargs)

    def get_fields(self):
        return self.core.get_bulk_change_fields(self.request) if self.fields is None else self.fields

    def get_fieldsets(self):
        return (
            (None, {'fields': self.get_fields()}),
        )

    def generate_fieldsets(self, form):
        return get_fieldsets_without_disallowed_fields(
            self.request,
            self.get_fieldsets(),
            (
                self._get_disallowed_fields_from_permissions()
                | set(self.generate_readonly_fields())
                | set(self.get_exclude())
            )
        )

    def get_readonly_fields(self):
        return ()

    def get_is_bulk(self):
        return True



import re
import json
import types
import datetime

from django.core.exceptions import ImproperlyConfigured
from django.contrib.admin.utils import display_for_value as admin_display_for_value
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import QuerySet
from django.core.exceptions import FieldDoesNotExist
from django.utils.translation import ugettext
from django.utils.html import format_html, format_html_join
from django.utils.formats import get_format, date_format
from django.utils.timezone import template_localtime

from chamber.utils import call_function_with_unknown_input

from pyston.converters import get_converter


PK_PATTERN = r'(?P<pk>[^/]+)'
NUMBER_PK_PATTERN = r'(?P<pk>\d+)'

EMPTY_VALUE = '---'

LOOKUP_SEP = '__'

METHOD_OBJ_STR_NAME = '_obj_name'


def is_callable(val):
    return hasattr(val, '__call__')


def get_new_class_name(prefix, klass):
    prefix = prefix.replace('-', ' ').title()
    prefix = re.sub(r'\s+', '', prefix)
    return prefix + klass.__name__


def flatten_fieldsets(fieldsets):
    """Returns a list of field names from an admin fieldsets structure."""
    field_names = []
    for _, opts in fieldsets or ():
        if 'fieldsets' in opts:
            field_names += flatten_fieldsets(opts.get('fieldsets'))
        else:
            for field in opts.get('fields', ()):
                if isinstance(field, (list, tuple)):
                    field_names.extend(field)
                else:
                    field_names.append(field)
    return field_names


def get_fieldsets_without_disallowed_fields(request, fieldsets, disallowed_fields):
    generated_fieldsets = []

    for title, fieldset_values in fieldsets:
        fieldset_values = dict(fieldset_values)
        if 'fields' in fieldset_values:
            fieldset_values['fields'] = [
                field for field in fieldset_values.pop('fields')
                if field not in disallowed_fields
            ]
        if 'fieldsets' in fieldset_values:
            fieldsets = get_fieldsets_without_disallowed_fields(
                request, fieldset_values.pop('fieldsets'), disallowed_fields
            )
            if fieldsets:
                fieldset_values['fieldsets'] = fieldsets
        if set(fieldset_values.keys()) & {'fields', 'fieldsets', 'inline_view_inst'}:
            generated_fieldsets.append((title, fieldset_values))
    return generated_fieldsets


def get_inline_views_from_fieldsets(fieldsets):
    """Returns a list of field names from an admin fieldsets structure."""
    inline_views = []
    for _, opts in fieldsets or ():
        if 'fieldsets' in opts:
            inline_views += get_inline_views_from_fieldsets(opts.get('fieldsets'))
        elif 'inline_view_inst' in opts:
            inline_views.append(opts.get('inline_view_inst'))
    return inline_views


def get_inline_views_opts_from_fieldsets(fieldsets):
    """Returns a list of field names from an admin fieldsets structure."""
    inline_views = []
    for _, opts in fieldsets or ():
        if 'fieldsets' in opts:
            inline_views += get_inline_views_opts_from_fieldsets(opts.get('fieldsets'))
        elif 'inline_view' in opts:
            inline_views.append(opts)
    return inline_views


def get_field_from_model_or_none(model, field_name):
    """
    Return field from model. If field doesn't exists null is returned instead of exception.
    """
    try:
        return model._meta.get_field(field_name)
    except (FieldDoesNotExist, AttributeError):
        return None


def get_field_label_from_path(model, field_path, view=None, field_labels=None):
    """
    Return field label of model class for input field path. For every field name in the field path is firstly get the
    right label and these labels are joined with " - " separator to one string.

    field_label input parameter can affect the result value. Example:

    * field_path='user__email', field_labels={} => 'user email'  # default values get from fields
    * field_path='user__email', field_labels={'user__email': 'e-mail'} => 'e-mail'  # full value is replaced
    * field_path='user__email', field_labels={'user': 'customer'} => 'customer - email'  # related field prefix is
        changed
    * field_path='user', field_labels={'user': 'customer'} => 'customer'  # full value is replaced
    * field_path='user', field_labels={'user__': 'customer'} => 'user'  # has no effect
    * field_path='user__email', field_labels={'user__': 'customer'} => 'customer email'  # related field prefix is
        changed
    * field_path='user__email', field_labels={'user__': None} => 'email'  # related field prefix is ignored
    * field_path='user__email', field_labels={'email': 'e-mail'} => 'user email'  # has no effect

    :param model: Django model class
    :param field_path: field names separated with "__"
    :param view: view instance
    :param field_labels: dict of field labels which can override result field name
    :return: field label
    """
    from .field_api import get_field_descriptors_from_path

    field_labels = {} if field_labels is None else field_labels

    field_descriptors = get_field_descriptors_from_path(model, field_path, view)

    used_field_names = []
    field_descriptor_labels = []

    for field_descriptor in field_descriptors:
        field_path_prefix = LOOKUP_SEP.join(used_field_names)
        current_field_path = LOOKUP_SEP.join(used_field_names + [field_descriptor.field_name])

        if field_descriptor_labels and field_path_prefix + LOOKUP_SEP in field_labels:
            if field_labels[field_path_prefix + LOOKUP_SEP] is not None:
                field_descriptor_labels = [field_labels[field_path_prefix + LOOKUP_SEP]]
            else:
                field_descriptor_labels = []

        if current_field_path in field_labels:
            if field_labels[current_field_path] is not None:
                field_descriptor_labels = [field_labels[current_field_path]]
            else:
                field_descriptor_labels = []
        elif field_descriptor.field_name != METHOD_OBJ_STR_NAME or not field_descriptor_labels:
            if field_descriptor.get_label() is not None:
                field_descriptor_labels.append(field_descriptor.get_label())

        used_field_names.append(field_descriptor.field_name)

    return ' - '.join([str(label) for label in field_descriptor_labels if label is not None])


def get_field_widget_from_path(model, field_path, view=None):
    """
    Return form widget to show value get from model instance and field_path
    """
    from .field_api import get_field_descriptors_from_path

    return get_field_descriptors_from_path(model, field_path, view)[-1].get_widget()


def get_readonly_field_value_from_path(instance, field_path, request=None, view=None):
    """
    Return ReadonlyValue instance which contains value and humanized value get from model instance and field_path
    """

    from .field_api import get_field_value_from_path

    return get_field_value_from_path(instance, field_path, request, view, return_readonly_value=True)


def get_readonly_field_data(instance, field_name, request, view=None, field_labels=None):
    """
    Returns field humanized value, label and widget which are used to display of instance or view readonly data.
    Args:
        field_name: name of the field which will be displayed
        instance: model instance
        view: view instance
        field_labels: dict of field labels which rewrites the generated field label

    Returns:
        field humanized value, label and widget which are used to display readonly data
    """
    return (
        get_readonly_field_value_from_path(instance, field_name, request, view),
        get_field_label_from_path(instance.__class__, field_name, view, field_labels),
        get_field_widget_from_path(instance.__class__, field_name, view)
    )


def display_object_data(obj, field_name, request, view=None):
    """
    Returns humanized value of model object that can be rendered to HTML or returned as part of REST

    examples:
       boolean True/False ==> Yes/No
       objects ==> object display name with link if current user has permissions to see the object
       field with choices ==> string value of choice
       field with humanize function ==> result of humanize function
    """
    return display_for_value(get_readonly_field_value_from_path(obj, field_name, request, view), request=request)


def display_code(value):
    """
    Display input value as a code.
    """
    return format_html(
        '<pre style="max-height: 400px">{}</pre>',
        value
    ) if value else display_for_value(value)


def display_json(value):
    """
    Display input JSON as a code
    """
    if value is None:
        return display_for_value(value)
    if isinstance(value, str):
        value = json.loads(value)
    return display_code(json.dumps(value, indent=2, ensure_ascii=False, cls=DjangoJSONEncoder))


def display_for_value(value, request=None):
    """
    Converts humanized value

    examples:
        boolean True/False ==> Yes/No
        objects ==> object display name with link if current user has permissions to see the object
        datetime ==> in localized format
        list ==> values separated with ","
        dict ==> string formatted with HTML ul/li tags
    """
    from is_core.forms.utils import ReadonlyValue
    from is_core.site import registered_model_cores

    if isinstance(value, ReadonlyValue):
        value = value.value

    if request and value.__class__ in registered_model_cores:
        return render_model_object_with_link(request, value)
    elif isinstance(value, (QuerySet, list, tuple, set, types.GeneratorType)):
        return format_html(
            '<ol class="field-list">{}</ol>',
            format_html_join(
                '\n',
                '<li>{}</li>',
                (
                    (display_for_value(v, request),) for v in value
                )
            )
        )
    elif isinstance(value, dict):
        return format_html(
            '<ul class="field-dict">{}</ul>',
            format_html_join(
                '\n',
                '{}{}',
                (
                    (
                        format_html('<li>{}</li>', k),
                        (
                            display_for_value(v, request) if isinstance(v, dict)
                            else format_html(
                                '<ul class="field-dict"><li>{}</li></ul>',
                                display_for_value(v, request)
                            )
                        )
                    )
                    for k, v in value.items()
                )
            )
        )
    elif isinstance(value, bool):
        return ugettext('Yes') if value else ugettext('No')
    elif isinstance(value, datetime.datetime):
        return date_format(template_localtime(value), (
             'DATETIME_FORMAT' if get_format('IS_CORE_VIEW_DATETIME_FORMAT') == 'IS_CORE_VIEW_DATETIME_FORMAT'
             else 'IS_CORE_VIEW_DATETIME_FORMAT'
         ))
    elif isinstance(value, datetime.date):
        return date_format(value, (
            'DATE_FORMAT' if get_format('IS_CORE_VIEW_DATE_FORMAT') == 'IS_CORE_VIEW_DATE_FORMAT'
            else 'IS_CORE_VIEW_DATE_FORMAT'
        ))
    else:
        return admin_display_for_value(value, EMPTY_VALUE)


def get_url_from_model_core(request, obj):
    """
    Returns object URL from model core.
    """
    from is_core.site import get_model_core
    model_core = get_model_core(obj.__class__)

    if model_core and hasattr(model_core, 'ui_patterns'):
        edit_pattern = model_core.ui_patterns.get('detail')
        return (
            edit_pattern.get_url_string(request, obj=obj)
            if edit_pattern and edit_pattern.has_permission('get', request, obj=obj) else None
        )
    else:
        return None


def get_obj_url(request, obj):
    """
    Returns object URL if current logged user has permissions to see the object
    """
    if (is_callable(getattr(obj, 'get_absolute_url', None)) and
            (not hasattr(obj, 'can_see_edit_link') or
             (is_callable(getattr(obj, 'can_see_edit_link', None)) and obj.can_see_edit_link(request)))):
        return call_function_with_unknown_input(obj.get_absolute_url, request=request)
    else:
        return get_url_from_model_core(request, obj)


def render_model_object_with_link(request, obj, display_value=None):
    if obj is None:
        return '[{}]'.format(ugettext('missing object'))

    obj_url = get_obj_url(request, obj)
    display_value = str(obj) if display_value is None else str(display_value)

    return format_html('<a href="{}">{}</a>', obj_url, display_value) if obj_url else display_value


def render_model_objects_with_link(request, objs):
    return format_html_join(', ', '{}', ((render_model_object_with_link(request, obj),) for obj in objs))


def header_name_to_django(header_name):
    return '_'.join(('HTTP', header_name.replace('-', '_').upper()))


def pretty_class_name(class_name):
    return re.sub(r'(\w)([A-Z])', r'\1-\2', class_name).lower()


def get_export_types_with_content_type(export_types):
    generated_export_types = []
    for title, type, serialization_format in export_types:
        try:
            generated_export_types.append(
                (title, type, serialization_format, get_converter(type).media_type)
            )
        except KeyError:
            raise ImproperlyConfigured('Missing converter for type {}'.format(type))
    return generated_export_types


def get_link_or_none(pattern_name, request, view_kwargs=None):
    """
    Helper that generate URL prom pattern name and kwargs and check if current request has permission to open the URL.
    If not None is returned.

    Args:
        pattern_name (str): slug which is used for view registratin to pattern
        request (django.http.request.HttpRequest): Django request object
        view_kwargs (dict): list of kwargs necessary for URL generator

    Returns:

    """
    from is_core.patterns import reverse_pattern

    pattern = reverse_pattern(pattern_name)
    assert pattern is not None, 'Invalid pattern name {}'.format(pattern_name)

    if pattern.has_permission('get', request, view_kwargs=view_kwargs):
        return pattern.get_url_string(request, view_kwargs=view_kwargs)
    else:
        return None


class GetMethodFieldMixin:

    @classmethod
    def get_method_returning_field_value(cls, field_name):
        """
        Method should return object method that can be used to get field value.
        Args:
            field_name: name of the field

        Returns: method for obtaining a field value

        """
        method = getattr(cls, field_name, None)
        return method if method and callable(method) else None


def get_model_name(model):
    return str(model._meta.model_name)



from django.utils.encoding import force_text

from pyston.utils import RestFieldset, RestField

from is_core.utils.field_api import get_field_descriptors_from_path, FieldIsNotRelation


def generate_subfield(submodel, subfield_name, field_class):
    if subfield_name and submodel:
        return field_class.create_from_string(subfield_name, submodel)
    elif submodel:
        return field_class('_obj_name')


class ModelRestFieldMixin:

    @classmethod
    def create_from_string(cls, full_field_name, model):
        subfield_name = None
        if '__' in full_field_name:
            full_field_name, subfield_name = full_field_name.split('__', 1)

        try:
            submodel = get_field_descriptors_from_path(model, full_field_name)[-1].get_related_model()
        except FieldIsNotRelation:
            submodel = None

        return cls(full_field_name, generate_subfield(submodel, subfield_name, cls))


class ModelRestField(ModelRestFieldMixin, RestField):

    def __init__(self, name, subfield=None):
        if isinstance(subfield, type(self)):
            subfield = ModelRestFieldset(subfield)
        super().__init__(name, subfield)


class ModelRestFlatField(ModelRestFieldMixin):

    def __init__(self, name, subfield=None):
        assert isinstance(name, str)
        assert subfield is None or isinstance(subfield, ModelRestFlatField)

        self.name = name
        self.subfield = subfield

    def __str__(self):
        if self.subfield:
            return '%s__%s' % (self.name, self.subfield)
        return '%s' % self.name


class ModelFlatRestFieldsMixin:

    @classmethod
    def create_from_flat_list(cls, fields_list, model):
        return cls(*[cls.fields_class.create_from_string(full_field_name, model)
                     for full_field_name in fields_list])


class ModelFlatRestFields(ModelFlatRestFieldsMixin):

    fields_class = ModelRestFlatField

    def __init__(self, *fields):
        self.fields = fields

    def __str__(self):
        return ','.join(map(force_text, self.fields))


class ModelRestFieldset(RestFieldset, ModelFlatRestFieldsMixin):

    fields_class = ModelRestField



import warnings
import itertools

from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.forms import models
from django.forms.fields import ChoiceField
from django.forms.models import ModelForm, _get_foreign_key, BaseModelFormSet

from pyston.forms import RestModelForm, RestFormMetaclass

from is_core.forms import widgets
from is_core.utils.models import get_model_field_value
from is_core.forms.formsets import BaseFormSetMixin, smartformset_factory
from is_core.forms.forms import SmartFormMixin


class BaseInlineFormSet(BaseFormSetMixin, models.BaseInlineFormSet):

    def _pre_save(self, obj):
        pass

    def _post_save_form(self, form):
        pass

    def _post_save(self, obj):
        pass

    def _set_post_save(self, commit, obj):
        if commit:
            self._post_save(obj)
        else:
            self.old_save_m2m = self.save_m2m

            def post_save_m2m():
                self.old_save_m2m()
                for form in self.saved_forms:
                    self._post_save_form(form)
                self._post_save(obj)
            self.save_m2m = post_save_m2m

    def save(self, commit=True):
        obj = super(BaseInlineFormSet, self).save(commit=False)
        self._pre_save(obj)
        if commit:
            obj.save()
        self._set_post_save(commit, obj)
        return obj


class ModelChoice(list):

    def __init__(self, id, label, obj=None, attrs={}):
        self.append(id)
        self.append(label)
        self.attrs = attrs
        self.obj = obj


class ModelChoiceIterator(forms.models.ModelChoiceIterator):

    def __iter__(self):
        return (
            ModelChoice(*choice) if not isinstance(choice, ModelChoice) else choice
            for choice in super(ModelChoiceIterator, self).__iter__()
        )

    def choice(self, obj):
        attrs = {}
        for key, val in obj._ui_meta.extra_selecbox_fields.items():
            attrs[key] = get_model_field_value(val, obj)
        return ModelChoice(self.field.prepare_value(obj), self.field.label_from_instance(obj), obj, attrs)

    def get_choice_from_value(self, value):
        try:
            obj = self.field.get_obj_from_value(value)
        except forms.ValidationError:
            return None

        if obj and self.queryset.filter(pk=obj.pk).exists():
            return self.choice(obj)
        else:
            return None


class ModelChoiceFieldMixin:

    widget = widgets.FulltextSelect

    def __init__(self, queryset, *args, **kwargs):
        self.model = queryset.model
        super(ModelChoiceFieldMixin, self).__init__(queryset, *args, **kwargs)

    def _get_choices(self):
        if hasattr(self, '_choices'):
            return self._choices
        return ModelChoiceIterator(self)
    choices = property(_get_choices, ChoiceField._set_choices)

    def widget_attrs(self, widget):
        attrs = super(ModelChoiceFieldMixin, self).widget_attrs(widget)
        options = self.model._meta
        attrs['data-model'] = '%s.%s' % (options.app_label, options.object_name)
        return attrs

    def get_obj_from_value(self, value):
        raise NotImplementedError


class ModelChoiceField(ModelChoiceFieldMixin, forms.ModelChoiceField):

    widget = widgets.RestrictedSelectWidget
    readonly_widget = widgets.ModelChoiceReadonlyWidget

    def get_obj_from_value(self, value):
        return self.to_python(value)


class ModelMultipleChoiceField(ModelChoiceFieldMixin, forms.ModelMultipleChoiceField):

    widget = widgets.RestrictedSelectMultipleWidget
    readonly_widget = widgets.ModelMultipleReadonlyWidget

    def get_obj_from_value(self, value):
        return self.to_python([value])[0]


class SmartModelFormMetaclass(RestFormMetaclass):

    def __new__(cls, name, bases, attrs):
        new_class = super(SmartModelFormMetaclass, cls).__new__(cls, name, bases, attrs)

        base_readonly_fields = set(getattr(new_class, 'base_readonly_fields', ()))
        base_required_fields = set(getattr(new_class, 'base_required_fields', ()))

        opts = getattr(new_class, 'Meta', None)
        if opts:
            setattr(opts, 'fields_to_clean', getattr(opts, 'fields_to_clean', None))
            exclude_fields = getattr(opts, 'exclude', None) or ()
            readonly_fields = getattr(opts, 'readonly_fields', None) or ()
            readonly = getattr(opts, 'readonly', None) or False
            all_fields = getattr(opts, 'all_fields', None) or ()
            has_all_fields = hasattr(opts, 'all_fields')
            required_fields = getattr(opts, 'required_fields', None) or ()

            for field_name, field in list(new_class.base_fields.items()):
                if field_name in exclude_fields:
                    del new_class.base_fields[field_name]
                elif ((field_name in readonly_fields or readonly or field.is_readonly) and
                      field_name not in base_readonly_fields):
                    base_readonly_fields.add(field_name)
                elif field_name in required_fields and field_name not in base_required_fields:
                    base_required_fields.add(field_name)

            if has_all_fields:
                test_readonly_fields = all_fields
            else:
                test_readonly_fields = readonly_fields

            for field_name in test_readonly_fields:
                if (field_name not in new_class.base_fields and 'formreadonlyfield_callback' in attrs and
                        attrs['formreadonlyfield_callback'] is not None):
                    new_class.base_fields[field_name] = attrs['formreadonlyfield_callback'](field_name)
                    base_readonly_fields.add(field_name)

        new_class.base_readonly_fields = base_readonly_fields
        new_class.base_required_fields = base_required_fields
        return new_class


def humanized_model_to_dict(instance, readonly_fields, fields=None, exclude=None):
    """
    Returns a dict containing the humanized data in ``instance`` suitable for passing as
    a Form's ``initial`` keyword argument.

    ``fields`` is an optional list of field names. If provided, only the named
    fields will be included in the returned dict.

    ``exclude`` is an optional list of field names. If provided, the named
    fields will be excluded from the returned dict, even if they are listed in
    the ``fields`` argument.
    """
    opts = instance._meta
    data = {}
    for f in itertools.chain(opts.concrete_fields, opts.private_fields, opts.many_to_many):
        if not getattr(f, 'editable', False):
            continue
        if fields and f.name not in fields:
            continue
        if f.name not in readonly_fields:
            continue
        if exclude and f.name in exclude:
            continue

        if f.humanized:
            data[f.name] = f.humanized(getattr(instance, f.name), instance)
    return data


class SmartModelForm(SmartFormMixin, RestModelForm, metaclass=SmartModelFormMetaclass):

    def __init__(self, *args, **kwargs):
        # Set values must be omitted
        readonly_exclude = kwargs.get('initials', {}).keys()

        super(SmartModelForm, self).__init__(*args, **kwargs)

        opts = self._meta
        self.humanized_data = humanized_model_to_dict(self.instance, self.readonly_fields, opts.fields,
                                                      itertools.chain(opts.exclude or (), readonly_exclude or ()))

    def _get_validation_exclusions(self):
        exclude = super(SmartModelForm, self)._get_validation_exclusions()
        for f in self.instance._meta.fields:
            field = f.name
            if field in self.readonly_fields and field not in exclude:
                exclude.append(field)
        return exclude


def get_model_fields(model, fields):
    model_fields = []
    for field in model._meta.fields + model._meta.many_to_many:
        if field.editable and (fields is None or field.name in fields):
            model_fields.append(field.name)
    return model_fields


def smartmodelform_factory(model, request, form=SmartModelForm, fields=None, readonly_fields=None, exclude=None,
                           formfield_callback=None, widgets=None, localized_fields=None, required_fields=None,
                           labels=None, help_texts=None, error_messages=None, field_classes=None,
                           formreadonlyfield_callback=None, readonly=False, fields_to_clean=None, is_bulk=False):
    attrs = {'model': model}
    if fields is not None:
        model_fields = get_model_fields(model, fields)
        attrs['all_fields'] = fields
        attrs['fields'] = model_fields
        attrs['non_model_fields'] = set(fields) - set(model_fields)
    if exclude is not None:
        attrs['exclude'] = exclude
    if widgets is not None:
        attrs['widgets'] = widgets
    if localized_fields is not None:
        attrs['localized_fields'] = localized_fields
    if labels is not None:
        attrs['labels'] = labels
    if help_texts is not None:
        attrs['help_texts'] = help_texts
    if error_messages is not None:
        attrs['error_messages'] = error_messages
    if field_classes is not None:
        attrs['field_classes'] = field_classes
    if readonly_fields is not None:
        attrs['readonly_fields'] = readonly_fields
    if required_fields is not None:
        attrs['required_fields'] = required_fields
    if fields_to_clean is not None:
        attrs['fields_to_clean'] = fields_to_clean
    attrs['readonly'] = readonly
    attrs['is_bulk'] = is_bulk

    bases = (form.Meta,) if hasattr(form, 'Meta') else ()
    Meta = type('Meta', bases, attrs)
    if formfield_callback:
        Meta.formfield_callback = staticmethod(formfield_callback)
        Meta.formreadonlyfield_callback = staticmethod(formreadonlyfield_callback)

    class_name = model.__name__ + str('Form')

    form_class_attrs = {
        'Meta': Meta,
        'formfield_callback': formfield_callback,
        'formreadonlyfield_callback': formreadonlyfield_callback,
        '_request': request,
    }

    if (getattr(Meta, 'fields', None) is None and
            getattr(Meta, 'exclude', None) is None):
        raise ImproperlyConfigured(
            "Calling modelform_factory without defining 'fields' or "
            "'exclude' explicitly is prohibited."
        )
    return type(form)(class_name, (form,), form_class_attrs)


def smartmodelformset_factory(model, request, form=ModelForm, formfield_callback=None,
                              formset=BaseModelFormSet, extra=1, can_delete=False,
                              can_order=False, max_num=None, fields=None, exclude=None,
                              widgets=None, validate_max=False, localized_fields=None,
                              labels=None, help_texts=None, error_messages=None,
                              min_num=None, validate_min=False, field_classes=None,
                              absolute_max=None, can_delete_extra=True, formreadonlyfield_callback=None,
                              readonly_fields=None, readonly=False):

    meta = getattr(form, 'Meta', None)
    if meta is None:
        meta = type(str('Meta'), (object,), {})
    if getattr(meta, 'fields', fields) is None and getattr(meta, 'exclude', exclude) is None:
        warnings.warn("Calling modelformset_factory without defining 'fields' or "
                      "'exclude' explicitly is deprecated",
                      PendingDeprecationWarning, stacklevel=2)

    form = smartmodelform_factory(
        model, request, form=form, fields=fields, exclude=exclude, formfield_callback=formfield_callback,
        widgets=widgets, localized_fields=localized_fields, labels=labels, help_texts=help_texts,
        error_messages=error_messages, formreadonlyfield_callback=formreadonlyfield_callback,
        readonly_fields=readonly_fields, readonly=readonly, field_classes=field_classes
    )
    FormSet = smartformset_factory(
        form, formset, extra=extra, min_num=min_num, max_num=max_num, can_order=can_order, can_delete=can_delete,
        validate_min=validate_min, validate_max=validate_max, absolute_max=absolute_max,
        can_delete_extra=can_delete_extra
    )
    FormSet.model = model
    return FormSet


def smartinlineformset_factory(parent_model, model, request, form=ModelForm,
                               formset=BaseInlineFormSet, fk_name=None,
                               fields=None, exclude=None, extra=3, can_order=False,
                               can_delete=True, min_num=None, max_num=None, formfield_callback=None,
                               widgets=None, validate_min=False, validate_max=False, localized_fields=None,
                               labels=None, help_texts=None, error_messages=None,
                               formreadonlyfield_callback=None, readonly_fields=None,
                               readonly=False):
    fk = _get_foreign_key(parent_model, model, fk_name=fk_name)
    # enforce a max_num=1 when the foreign key to the parent model is unique.
    if fk.unique:
        max_num = 1
    kwargs = {
        'form': form,
        'formfield_callback': formfield_callback,
        'formset': formset,
        'extra': extra,
        'can_delete': can_delete,
        'can_order': can_order,
        'fields': fields,
        'exclude': exclude,
        'max_num': max_num,
        'min_num': min_num,
        'widgets': widgets,
        'validate_min': validate_min,
        'validate_max': validate_max,
        'localized_fields': localized_fields,
        'labels': labels,
        'help_texts': help_texts,
        'error_messages': error_messages,
        'formreadonlyfield_callback': formreadonlyfield_callback,
        'readonly_fields': readonly_fields,
        'readonly': readonly,
    }
    FormSet = smartmodelformset_factory(model, request, **kwargs)
    FormSet.fk = fk
    return FormSet



from django.contrib.contenttypes.models import ContentType
from django.shortcuts import render
from django.utils import translation
from django.utils.encoding import force_text
from django.utils.translation import ugettext

from is_core.rest.resource import CoreResource, DjangoCoreResource

from django_celery_extensions.task import obj_to_string

from .models import ExportedFile
from .tasks import background_serialization


class ErrorResponseData(dict):

    def __init__(self, msg, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self['messages'] = {'error': force_text(msg)}


def apply_background_export(user, queryset, rest_context, fieldset, serialization_format, filename):
    exported_file = ExportedFile.objects.create(
        created_by=user,
        content_type=ContentType.objects.get_for_model(queryset.model)
    )

    background_serialization.apply_async_on_commit(
        args=(
            exported_file.pk,
            rest_context,
            translation.get_language(),
            fieldset,
            serialization_format,
            filename,
            obj_to_string(queryset.query),
        ),
        related_objects=[exported_file]
    )


class CeleryResourceMixin:

    def _get_error_response_data(self, message):
        return {
            'messages': {'error': force_text(message)}
        }

    def _get_no_background_permissions_response_data(self, http_headers):
        return ErrorResponseData(
            ugettext('User doesn\'t have permissions to export')
        ), http_headers, 403, False

    def _serialize_data_in_background(self, result):
        apply_background_export(
            self.request.user,
            result,
            self.request._rest_context,
            force_text(self._get_requested_fieldset(result)),
            self._get_serialization_format(),
            self._get_filename(),
        )

    def _get_filename(self):
        parts = super()._get_filename().rsplit(sep='.', maxsplit=1)
        return (
            '{}{}.{}'.format(parts[0], '_DUVERNE', parts[1]) if len(parts) == 2
            else '{}{}'.format(super()._get_filename(), '_DUVERNE')
        )

    def _get_background_serialization_response_data(self, result, http_headers):
        if not self.permission.has_permission('export', self.request, self):
            return self._get_no_background_permissions_response_data(http_headers)
        else:
            self._serialize_data_in_background(result)
            response = render(self.request, 'is_core/background_serialization.html')
            return response, response['Content-Type'], 200, False

    def _get_paginator(self):
        # For background serialization paginator is not used
        return None if 'background_serialization' in self.request._rest_context else self.paginator

    def _get_response_data(self):
        result, http_headers, status_code, fieldset = super()._get_response_data()
        if 'background_serialization' in self.request._rest_context:
            return self._get_background_serialization_response_data(result, http_headers)
        else:
            return result, http_headers, status_code, fieldset

    def _get_headers_queryset_context_mapping(self):
        context_mapping = super()._get_headers_queryset_context_mapping()
        context_mapping['background_serialization'] = ('BACKGROUND_SERIALIZATION', '_background_serialization')
        return context_mapping

    def _get_name(self):
        obj = self._get_obj_or_none(pk=self.kwargs.get('pk'))
        return force_text(obj).replace(' ', '-') if obj else super()._get_name()


class CeleryDjangoCoreResource(CeleryResourceMixin, DjangoCoreResource):
    pass


class CeleryCoreResource(CeleryResourceMixin, CoreResource):
    pass



from django.utils.translation import ugettext_lazy as _

from is_core.auth.permissions import PermissionsSet, CoreCreateAllowed, CoreAllowed, DEFAULT_PERMISSION

from .form_views import DjangoCoreFormView


class DjangoAddFormView(DjangoCoreFormView):

    template_name = 'is_core/generic_views/add_form.html'
    form_template = 'is_core/forms/model_add_form.html'
    view_type = 'add'
    messages = {'success': _('The %(name)s "%(obj)s" was added successfully.'),
                'error': _('Please correct the error below.')}

    permission = PermissionsSet(
        get=CoreCreateAllowed(),
        post=CoreCreateAllowed(),
        **{
            DEFAULT_PERMISSION: CoreAllowed(),
        }
    )

    def get_view_title(self):
        return (self.title or
                self.model._ui_meta.add_verbose_name % {'verbose_name': self.model._meta.verbose_name,
                                                        'verbose_name_plural': self.model._meta.verbose_name_plural})



from is_core.generic_views.inlines.inline_table_views import BaseModelInlineTableView
from is_core.generic_views.table_views import BaseModelTableView
from is_core.generic_views.detail_views import ModelReadonlyDetailView


class DynamoTableViewMixin:

    list_verbose_name = '%(verbose_name_plural)s'

    @property
    def model_name(self):
        return str(self.model.__class__.__name__.lower())


class DynamoInlineTableView(DynamoTableViewMixin, BaseModelInlineTableView):
    pass


class DynamoTableView(DynamoTableViewMixin, BaseModelTableView):

    add_button_verbose_name = 'add %(verbose_name)s'


class DynamoDetailView(ModelReadonlyDetailView):

    detail_verbose_name = '%(verbose_name)s'

    def _get_obj_or_none(self):
        return self.core.get_obj(self.request, self.kwargs['pk'])



from django.conf import settings

from elasticsearch_dsl import Document, Date, Integer, Keyword, Text, Boolean
from elasticsearch_dsl import connections


connections.create_connection(**settings.ELASTICSEARCH_DATABASE)


class Comment(Document):

    user_id = Keyword()
    content = Text()
    is_public = Boolean()
    priority = Integer()

    @property
    def id(self):
        return self.meta.id

    @property
    def pk(self):
        return self.meta.id

    def __str__(self):
        return self.id

    class Index:
        name = 'comment'


Comment.init()



from germanium.decorators import login

from germanium.test_cases.rest import RestTestCase
from germanium.tools.trivials import assert_in, assert_equal, assert_true, assert_is_not_none
from germanium.tools.http import (assert_http_bad_request, assert_http_not_found, assert_http_method_not_allowed,
                                  assert_http_accepted, build_url, assert_http_unauthorized, assert_http_ok,
                                  assert_http_redirect)
from germanium.tools.rest import assert_valid_JSON_created_response, assert_valid_JSON_response

from .test_case import HelperTestCase, AsSuperuserTestCase

from issue_tracker.dynamo.models import Comment


class RESTDynamodDBTestCase(AsSuperuserTestCase, HelperTestCase, RestTestCase):

    COMMENT_API_URL = '/api/dynamo-comment/{}/'
    COMMENT_UI_URL = '/dynamo-comment/{}/'

    @classmethod
    def setUpClass(cls):
        Comment.create_table()
        for i in range(10):
            comment = Comment(
                user_id=str(i),
                issue_id=str(i % 3),
                content=f'test message {i}',
                is_public=bool(i % 2),
                priority=i
            )
            comment.save()

    @classmethod
    def tearDownClass(cls):
        Comment.delete_table()

    @login(is_superuser=True)
    def test_get_api_dynamo_comments_should_return_right_data(self):
        resp = self.get(f'{self.COMMENT_API_URL.format(0)}?_fields=user_id,content,is_public,priority')
        assert_valid_JSON_response(resp)
        assert_equal(len(resp.json()), 4)
        for i, data in enumerate(resp.json()):
            assert_equal(data['priority'], i * 3)
            assert_equal(data['user_id'], str(i * 3))
            assert_equal(data['is_public'], bool(i % 2))
            assert_equal(data['content'], f'test message {i * 3}')

    @login(is_superuser=True)
    def test_get_api_dynamo_one_comment_should_return_right_data(self):
        resp = self.get(f'{self.COMMENT_API_URL.format(0)}3/?_fields=id,user_id,content,is_public,priority')
        assert_valid_JSON_response(resp)
        data = resp.json()
        assert_equal(data['priority'], 3)
        assert_equal(data['user_id'], '3')
        assert_equal(data['is_public'], True)
        assert_equal(data['content'], f'test message 3')

    def test_get_api_dynamo_comments_without_authorization_should_return_unauthorized(self):
        assert_http_unauthorized(self.get(self.COMMENT_API_URL.format(0)))

    @login(is_superuser=True)
    def test_get_dynamo_should_return_list_view(self):
        assert_http_ok(self.get(self.COMMENT_UI_URL.format(0)))

    @login(is_superuser=True)
    def test_get_dynamo_should_return_detail_view(self):
        assert_http_ok(self.get(f'{self.COMMENT_UI_URL.format(0)}3/'))

    @login(is_superuser=True)
    def test_get_dynamo_should_return_not_found_for_invalid_object_detail(self):
        assert_http_not_found(self.get(f'{self.COMMENT_UI_URL.format(1)}3/'))

    def test_get_dynamo_comments_without_authorization_should_return_redirect_to_login(self):
        assert_http_redirect(self.get(f'{self.COMMENT_UI_URL.format(0)}0/'))
        assert_http_redirect(self.get(self.COMMENT_UI_URL.format(0)))



from django import forms
from django.db.models.fields import DateField, DateTimeField
from django.db.models.fields.related import ForeignKey, ManyToManyField, ForeignObjectRel

from pyston.filters.filters import NONE_LABEL
from pyston.filters.utils import OperatorSlug
from pyston.filters.managers import DjangoFilterManager
from pyston.filters.django_filters import (
    DateFieldFilter, ManyToManyFieldFilter, ForeignKeyFieldFilter, ForeignObjectRelFilter
)

from is_core.forms.widgets import DateRangeFilterWidget, DateTimeRangeFilterWidget, RestrictedSelectWidget


class FilterChoiceIterator:
    """
    Filter iterator surrounds default model iterator, removes blank values and adds possibilities
    to remove filtered value (blank) and filter according to "None" value.
    """

    def __init__(self, choices, field=None):
        self.choices = choices
        self.field = field

    def __iter__(self):
        yield ('', '')

        if self.field and (self.field.null or self.field.blank):
            yield ('__none__', NONE_LABEL)

        for k, v in self.choices:
            if k is not None and k != '':
                yield (k, v)

    def __len__(self):
        return len(self.choices)

    def __getattr__(self, name):
        return getattr(self.choices, name)


class UIFilterMixin:
    """
    Special mixin for improve Pyston filters. There can be defined UI appearance (widget) and used operator.
    """

    widget = None

    def get_widget(self, request):
        """
        Returns concrete widget that will be used for rendering table filter.
        """
        widget = self.widget
        if isinstance(widget, type):
            widget = widget()
        return widget

    def get_operator(self, widget):
        """
        Returns operator used for filtering. By default it is first operator defined in Pyston filter.
        """
        return self.get_allowed_operators()[0]


class DateRangeFieldFilter(UIFilterMixin, DateFieldFilter):
    """
    UI filter for date field that provides possibility to filter date from/to
    """

    widget = DateRangeFilterWidget


class DateTimeRangeFieldFilter(UIFilterMixin, DateFieldFilter):
    """
    UI filter for datetime field that provides possibility to filter date from/to
    """

    widget = DateTimeRangeFilterWidget


class RelatedUIFilter(UIFilterMixin):
    """
    Helper for all filters of related field.
    """

    def _update_widget_choices(self, widget):
        """
        Updates widget choices with special choice iterator that removes blank values and adds none value to clear
        filter data.
        :param widget: widget with choices
        :return: updated widget with filter choices
        """

        widget.choices = FilterChoiceIterator(widget.choices, self.field)
        return widget

    def get_operator(self, widget):
        """
        Because related form field can be restricted (if choices is too much it is used textarea without select box)
        for situation without choices (textarea) is used contains operator.
        :param widget: restricted widget
        :return: operator that will be used for filtering
        """
        return OperatorSlug.CONTAINS if widget.is_restricted else OperatorSlug.EQ


class UIForeignKeyFieldFilter(RelatedUIFilter, ForeignKeyFieldFilter):

    def get_widget(self, request):
        """
        Field widget is replaced with "RestrictedSelectWidget" because we not want to use modified widgets for
        filtering.
        """
        return self._update_widget_choices(self.field.formfield(widget=RestrictedSelectWidget).widget)


class UIManyToManyFieldFilter(RelatedUIFilter, ManyToManyFieldFilter):

    def get_widget(self, request):
        """
        Field widget is replaced with "RestrictedSelectWidget" because "MultipleChoiceField" is not optional for
        filtering purposes.
        """
        return self._update_widget_choices(self.field.formfield(widget=RestrictedSelectWidget).widget)


class UIForeignObjectRelFilter(RelatedUIFilter, ForeignObjectRelFilter):

    def get_widget(self, request):
        """
        Table view is not able to get form field from reverse relation.
        Therefore this widget returns similar form field as direct relation (ModelChoiceField).
        Because there is used "RestrictedSelectWidget" it is returned textarea or selectox with choices according to
        count objects in the queryset.
        """
        return self._update_widget_choices(
            forms.ModelChoiceField(
                widget=RestrictedSelectWidget, queryset=self.field.related_model._default_manager.all()
            ).widget
        )


class CoreDjangoFilterManager(DjangoFilterManager):

    model_field_filters = {
        DateField: DateRangeFieldFilter,
        DateTimeField: DateTimeRangeFieldFilter,
        ManyToManyField: UIManyToManyFieldFilter,
        ForeignKey: UIForeignKeyFieldFilter,
        ForeignObjectRel: UIForeignObjectRelFilter,
    }



from pyston.contrib.elasticsearch.paginator import \
    ElasticsearchOffsetBasedPaginator as OriginElasticsearchOffsetBasedPaginator


class ElasticsearchOffsetBasedPaginator(OriginElasticsearchOffsetBasedPaginator):

    type = 'offset-based-paginator'

    def _get_total(self, qs, request):
        if request._rest_context.get('request_count', False):
            return super()._get_total(qs, request)
        else:
            return None



from is_core.auth.permissions import PermissionsSet, SelfPermission, IsSuperuser
from is_core.main import DjangoUiRestCore

from .models import ExportedFile
from .resource import CeleryDjangoCoreResource


class BackgroundExportCoreMixin:

    export_permission = IsSuperuser()
    rest_resource_class = CeleryDjangoCoreResource

    def _get_export_permission(self):
        return self.export_permission

    def _init_permission(self, permission):
        permission = super()._init_permission(permission)
        permission.set('export', self._get_export_permission())
        return permission


class DjangoBackgroundExportUiRestCore(BackgroundExportCoreMixin, DjangoUiRestCore):

    abstract = True


class BaseExportedFileCore(DjangoBackgroundExportUiRestCore):

    abstract = True
    model = ExportedFile

    can_create = can_update = can_delete = False

    read_own_permission = IsSuperuser()
    read_all_permission = IsSuperuser()

    all_list_fields = (
        'changed_at', 'created_at', 'created_by', 'downloaded_by', 'expiration', 'download_link'
    )
    own_list_fields = (
        'changed_at', 'created_at', 'expiration', 'download_link'
    )
    form_fields = (
        'changed_at', 'created_at', 'created_by', 'downloaded_by', 'content_type', 'download_link', 'expiration'
    )

    def _init_permission(self, permission):
        return PermissionsSet(
            read=SelfPermission('read_own') | SelfPermission('read_all'),
            read_own=self.read_own_permission,
            read_all=self.read_all_permission,
        )

    def get_list_fields(self, request):
        return (
            list(self.all_list_fields) if self.permission.has_permission('read_all', request, self)
            else list(self.own_list_fields)
        )

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        if self.permission.has_permission('read_all', request, self):
            return qs
        elif self.permission.has_permission('read_own', request, self):
            return qs.filter(created_by=request.user)
        else:
            return qs.none()



from .response import (ResponseException, HttpRedirectResponseException, HttpBadRequestResponseException,  # noqa: F401
                       HttpUnauthorizedResponseException, HttpForbiddenResponseException,
                       HttpUnsupportedMediaTypeResponseException, HttpMethodNotAllowedResponseException,
                       HttpDuplicateResponseException, HttpServiceUnavailableException,)

from chamber.exceptions import PersistenceException  # noqa: F401



from pyston.contrib.elasticsearch.filters import ElasticsearchFilterManager, DateTimeElasticsearchFilter

from is_core.rest.filters import UIFilterMixin
from is_core.forms.widgets import DateTimeRangeFilterWidget


class DateTimeRangeFieldFilter(UIFilterMixin, DateTimeElasticsearchFilter):

    widget = DateTimeRangeFilterWidget


class CoreElasticsearchFilterManagerFilterManager(ElasticsearchFilterManager):

    filter_by_field_name = {
        **ElasticsearchFilterManager.filter_by_field_name,
        'date': DateTimeRangeFieldFilter,
    }



from pyston.contrib.dynamo.paginator import DynamoCursorBasedPaginator as OriginDynamoCursorBasedPaginator


class DynamoCursorBasedPaginator(OriginDynamoCursorBasedPaginator):

    type = 'cursor-based-paginator'



from is_core.main import ModelCore, ModelUiCore, ModelUiRestCore, ModelRestCore

from .filters import CoreElasticsearchFilterManagerFilterManager
from .paginator import ElasticsearchOffsetBasedPaginator
from .resources import ElasticsearchCoreResource
from .views import ElasticsearchDetailView, ElasticsearchTableView


class ElasticsearchCore(ModelCore):

    abstract = True

    @property
    def menu_group(self):
        return self.model._index._name.replace('-', '_')

    def get_queryset(self, request):
        queryset = self.model.search()
        ordering = self.get_default_ordering()
        if ordering:
            queryset = queryset.sort(*ordering)
        return queryset


class ElasticsearchUiCore(ElasticsearchCore, ModelUiCore):

    abstract = True

    ui_list_view = ElasticsearchTableView
    ui_detail_view = ElasticsearchDetailView


class ElasticsearchRestCore(ElasticsearchCore, ModelRestCore):

    abstract = True

    rest_resource_class = ElasticsearchCoreResource
    rest_paginator = ElasticsearchOffsetBasedPaginator()
    rest_filter_manager = CoreElasticsearchFilterManagerFilterManager()


class ElasticsearchUiRestCore(ElasticsearchRestCore, ElasticsearchUiCore, ModelUiRestCore):

    abstract = True



from elasticsearch import NotFoundError

from is_core.rest.resource import ModelCoreResourceMixin

from pyston.contrib.elasticsearch.resource import BaseElasticsearchResource


class ElasticsearchCoreResource(ModelCoreResourceMixin, BaseElasticsearchResource):

    def _get_obj_or_none(self, pk=None):
        if not pk:
            return None
        try:
            return self.core.model.get(id=pk)
        except NotFoundError:
            return None



# Generated by Django 3.2 on 2022-12-06 12:51

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('issue_tracker', '0002_auto_20210401_1128'),
    ]

    operations = [
        migrations.AddField(
            model_name='issue',
            name='parent',
            field=models.ForeignKey(
                blank=True,
                null=True, on_delete=django.db.models.deletion.CASCADE,
                related_name='subissues',
                to='issue_tracker.issue',
                verbose_name='Parent'
            ),
        ),
    ]



from django.utils.translation import gettext_lazy as _l

from is_core.main import ModelCore, ModelUiCore, ModelUiRestCore, ModelRestCore
from is_core.utils.decorators import short_description

from .paginator import DynamoCursorBasedPaginator
from .resources import DynamoCoreResource
from .views import DynamoDetailView, DynamoTableView


class DynamoCore(ModelCore):

    abstract = True

    rest_paginator = DynamoCursorBasedPaginator()

    rest_range_key = None

    @property
    def menu_group(self):
        return self.model.__class__.__name__.lower()

    def get_queryset(self, request):
        return self.model.objects.set_hash_key(self._get_hash_key(request))

    def _get_hash_key(self, request):
        raise NotImplementedError

    def _get_obj(self, request, pk):
        return self.model.get(self._get_hash_key(request), pk)

    def get_obj(self, request, pk):
        if not pk:
            return None
        try:
            return self._get_obj(request, pk)
        except self.model.DoesNotExist:
            return None

    @short_description(_l('object name'))
    def _obj_name(self):
        return self.verbose_name


class DynamoUiCore(DynamoCore, ModelUiCore):

    abstract = True

    ui_list_view = DynamoTableView
    ui_detail_view = DynamoDetailView


class DynamoRestCore(DynamoCore, ModelRestCore):

    abstract = True

    rest_paginator = DynamoCursorBasedPaginator()
    rest_resource_class = DynamoCoreResource


class DynamoUiRestCore(DynamoUiCore, DynamoRestCore, ModelUiRestCore):

    abstract = True



from functools import wraps

from django.conf import settings as django_settings
from django.http.response import Http404
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext
from django.utils.encoding import force_text
from django.urls import NoReverseMatch

from pyston.conf import settings as pyston_settings
from pyston.forms import rest_modelform_factory
from pyston.exception import (RestException, MimerDataException, NotAllowedException, UnsupportedMediaTypeException,
                              ResourceNotFoundException, NotAllowedMethodException, DuplicateEntryException,
                              ConflictException, DataInvalidException, UnauthorizedException)
from pyston.resource import BaseResource, DjangoResource
from pyston.response import RestErrorResponse, RestErrorsResponse

from chamber.shortcuts import get_object_or_none
from chamber.utils import transaction

from is_core.auth.permissions import (
    PermissionsSet, IsAuthenticated, CoreAllowed, CoreReadAllowed, CoreUpdateAllowed, CoreDeleteAllowed,
    CoreCreateAllowed, AllowAny, DEFAULT_PERMISSION
)
from is_core.auth.views import FieldPermissionViewMixin
from is_core.exceptions.response import (HttpBadRequestResponseException, HttpUnsupportedMediaTypeResponseException,
                                         HttpMethodNotAllowedResponseException, HttpDuplicateResponseException,
                                         HttpForbiddenResponseException)
from is_core.forms.models import smartmodelform_factory
from is_core.patterns import RestPattern, patterns
from is_core.utils import get_field_label_from_path, METHOD_OBJ_STR_NAME, LOOKUP_SEP

from .filters import CoreDjangoFilterManager


class PermissionsResourceMixin:

    permission = IsAuthenticated()
    csrf_exempt = False

    def has_get_permission(self, **kwargs):
        return (
            self.permission.has_permission('get', self.request, self, obj=kwargs.get('obj')) and
            super().has_get_permission(**kwargs)
        )

    def has_post_permission(self, **kwargs):
        return (
            self.permission.has_permission('post', self.request, self, obj=kwargs.get('obj')) and
            super().has_post_permission(**kwargs)
        )

    def has_put_permission(self, **kwargs):
        return (
            self.permission.has_permission('put', self.request, self, obj=kwargs.get('obj')) and
            super().has_put_permission(**kwargs)
        )

    def has_patch_permission(self, **kwargs):
        return (
            self.permission.has_permission('patch', self.request, self, obj=kwargs.get('obj')) and
            super().has_patch_permission(**kwargs)
        )

    def has_delete_permission(self, **kwargs):
        return (
            self.permission.has_permission('delete', self.request, self, obj=kwargs.get('obj')) and
            super().has_delete_permission(**kwargs)
        )

    def has_options_permission(self, **kwargs):
        return (
            self._is_cors_options_request() or (
                self.permission.has_permission('options', self.request, self, obj=kwargs.get('obj')) and
                super().has_options_permission(**kwargs)
            )
        )

    def has_head_permission(self, **kwargs):
        return (
            self.permission.has_permission('head', self.request, self, obj=kwargs.get('obj')) and
            super().has_head_permission(**kwargs)
        )

    def _check_permission(self, name, *args, **kwargs):
        try:
            super()._check_permission(name, *args, **kwargs)
        except NotAllowedException:
            if not hasattr(self.request, 'user') or not self.request.user or not self.request.user.is_authenticated:
                raise UnauthorizedException
            else:
                raise


class PermissionsModelResourceMixin(PermissionsResourceMixin):

    can_create_obj = True
    can_read_obj = True
    can_update_obj = True
    can_delete_obj = True

    def has_create_obj_permission(self, obj=None, **kwargs):
        return (
            self.permission.has_permission('create_obj', self.request, self, obj=obj) and
            super().has_create_obj_permission(obj=obj, **kwargs)
        )

    def has_read_obj_permission(self, obj=None, **kwargs):
        return (
            self.permission.has_permission('read_obj', self.request, self, obj=obj) and
            super().has_read_obj_permission(obj=obj, **kwargs)
        )

    def has_update_obj_permission(self, obj=None, **kwargs):
        return (
            self.permission.has_permission('update_obj', self.request, self, obj=obj) and
            super().has_update_obj_permission(obj=obj, **kwargs)
        )

    def has_delete_obj_permission(self, obj=None, **kwargs):
        return (
            self.permission.has_permission('delete_obj', self.request, self, obj=obj) and
            super().has_delete_obj_permission(obj=obj, **kwargs)
        )


class CoreResourceMixin:

    register = False
    abstract = True

    def dispatch(self, request, *args, **kwargs):
        if hasattr(self, 'core'):
            self.core.init_rest_request(request)
        return super().dispatch(request, *args, **kwargs)

    @classmethod
    def __init_core__(cls, core, pattern):
        cls.core = core
        cls.pattern = pattern

    def _get_error_response(self, exception):
        """
        Trasform pyston exceptions to Is-core exceptions and raise it
        """
        response_exceptions = {
            MimerDataException: HttpBadRequestResponseException,
            NotAllowedException: HttpForbiddenResponseException,
            UnsupportedMediaTypeException: HttpUnsupportedMediaTypeResponseException,
            Http404: Http404,
            ResourceNotFoundException: Http404,
            NotAllowedMethodException: HttpMethodNotAllowedResponseException,
            DuplicateEntryException: HttpDuplicateResponseException,
            ConflictException: HttpDuplicateResponseException,
        }
        response_exception = response_exceptions.get(type(exception))
        if response_exception:
            raise response_exception
        return super()._get_error_response(exception)

    @classmethod
    def get_method_returning_field_value(cls, field_name):
        """
        Field values can be obtained from resource or core.
        """
        method = super().get_method_returning_field_value(field_name)

        if method:
            return method

        core_method = cls.core.get_method_returning_field_value(field_name)
        if core_method:
            @wraps(core_method)
            def core_method_wrapper(self, **kwargs):
                return core_method(self.core, **kwargs)
            return core_method_wrapper

        return None


class PermissionsModelCoreResourceMixin(PermissionsModelResourceMixin):

    pk_name = 'pk'

    permission = PermissionsSet(
        # HTTP permissions
        head=CoreReadAllowed(),
        options=CoreReadAllowed(),
        post=CoreCreateAllowed(),
        get=CoreReadAllowed(),
        put=CoreUpdateAllowed(),
        patch=CoreUpdateAllowed(),
        delete=CoreDeleteAllowed(),

        # Serializer permissions
        create_obj=CoreCreateAllowed(),
        read_obj=CoreReadAllowed(),
        update_obj=CoreUpdateAllowed(),
        delete_obj=CoreDeleteAllowed(),

        # Other permissions
        **{
            DEFAULT_PERMISSION: CoreAllowed(),
        }
    )

    def _get_perm_obj_or_none(self, pk=None):
        pk = pk or self.kwargs.get(self.pk_name)
        if pk:
            return get_object_or_none(self.core.model, pk=pk)
        else:
            return None


class CoreResource(PermissionsResourceMixin, CoreResourceMixin, BaseResource):
    pass


class EntryPointResource(CoreResource):

    allowed_methods = ('get', 'head', 'options')
    permission = AllowAny()

    def get(self):
        out = {}
        for pattern_name, pattern in patterns.items():
            if isinstance(pattern, RestPattern):
                try:
                    url = pattern.get_url_string(self.request)
                    allowed_methods = pattern.get_allowed_methods(self.request, None)
                    if allowed_methods:
                        out[pattern_name] = {'url': url, 'methods': allowed_methods}
                except NoReverseMatch:
                    pass

        return out


class SmartFormDjangoResource(DjangoResource):

    def _generate_form_class(self, inst, exclude=None):
        exclude = [] if exclude is None else exclude
        exclude = list(self._get_exclude(inst)) + exclude
        form_class = self._get_form_class(inst)
        fields = self._get_form_fields(inst)
        if hasattr(form_class, '_meta') and form_class._meta.exclude:
            exclude.extend(form_class._meta.exclude)
        return rest_modelform_factory(
            self.model,
            form=form_class,
            form_factory=smartmodelform_factory,
            resource_typemapper=self.resource_typemapper,
            auto_related_direct_fields=pyston_settings.AUTO_RELATED_DIRECT_FIELDS,
            auto_related_reverse_fields=pyston_settings.AUTO_RELATED_REVERSE_FIELDS,
            request=self.request,
            exclude=exclude,
            fields=fields,
            formfield_callback=self.formfield_for_dbfield,
            labels=self.get_field_labels()
        )


class ModelCoreResourceMixin(FieldPermissionViewMixin, CoreResourceMixin, PermissionsModelCoreResourceMixin):

    DEFAULT_REST_CONTEXT_MAPPING = {
        **BaseResource.DEFAULT_REST_CONTEXT_MAPPING,
        'request_count': ('HTTP_X_REQUEST_COUNT', '_request_count'),
        'direction': ('HTTP_X_DIRECTION', '_direction'),
        'order': ('HTTP_X_ORDER', '_order')
    }

    field_labels = None
    filters = {}

    def _get_queryset(self):
        return self.core.get_queryset(self.request)

    def get_allowed_fields_rfs(self, obj=None):
        return super().get_allowed_fields_rfs().subtract(self._get_disallowed_fields_from_permissions(obj=obj))

    def get_fields(self, obj=None):
        fields = list(self.fields) if self.fields is not None else None
        return self.core.get_rest_fields(self.request, obj=None) if fields is None else fields

    def get_default_fields(self, obj=None):
        default_fields = list(self.default_fields) if self.default_fields is not None else None
        return self.core.get_rest_default_fields(self.request, obj=None) if default_fields is None else default_fields

    def get_detailed_fields(self, obj=None):
        detailed_fields = list(self.detailed_fields) if self.detailed_fields is not None else self.get_fields(obj=obj)
        return self.core.get_rest_detailed_fields(self.request, obj=obj) if detailed_fields is None else detailed_fields

    def get_general_fields(self, obj=None):
        general_fields = list(self.general_fields) if self.general_fields is not None else self.get_fields(obj=obj)
        return self.core.get_rest_general_fields(self.request, obj=obj) if general_fields is None else general_fields

    def get_guest_fields(self, obj=None):
        guest_fields = list(self.guest_fields) if self.guest_fields is not None else None
        return self.core.get_rest_guest_fields(self.request, obj=obj) if guest_fields is None else guest_fields

    def get_extra_fields(self, obj=None):
        extra_fields = list(self.extra_fields) if self.extra_fields is not None else None
        return self.core.get_rest_extra_fields(self.request) if extra_fields is None else extra_fields

    def get_extra_filter_fields(self):
        extra_filter_fields = list(self.extra_filter_fields) if self.extra_filter_fields is not None else None
        return (
            self.core.get_rest_extra_filter_fields(self.request) if extra_filter_fields is None else extra_filter_fields
        )

    def get_filter_fields(self):
        filter_fields = list(self.filter_fields) if self.filter_fields is not None else None
        return self.core.get_rest_filter_fields(self.request) if filter_fields is None else filter_fields

    def get_extra_order_fields(self):
        extra_order_fields = list(self.extra_order_fields) if self.extra_order_fields is not None else None
        return self.core.get_rest_extra_order_fields(self.request) if extra_order_fields is None else extra_order_fields

    def get_order_fields(self):
        order_fields = list(self.order_fields) if self.order_fields is not None else None
        return self.core.get_rest_order_fields(self.request) if order_fields is None else order_fields

    def get_field_labels(self):
        return (
            self.field_labels if self.field_labels is not None else self.core.get_rest_field_labels(self.request)
        )

    def get_field_label(self, field_name):
        return get_field_label_from_path(
            self.model,
            (
                field_name[:-len(LOOKUP_SEP + METHOD_OBJ_STR_NAME)]
                if field_name.endswith(LOOKUP_SEP + METHOD_OBJ_STR_NAME)
                else field_name
            ),
            field_labels=self.get_field_labels()
        )

    def _get_exclude(self, obj=None):
        return (
            list(self.core.get_rest_form_exclude(self.request, obj))
            + list(self._get_readonly_fields_from_permissions(obj))
        )


class DjangoCoreResource(ModelCoreResourceMixin, SmartFormDjangoResource):

    form_class = None
    abstract = True
    filter_manager = CoreDjangoFilterManager()

    def _preload_queryset(self, qs):
        return self.core.preload_queryset(self.request, qs)

    def _get_form_fields(self, obj=None):
        return self.core.get_rest_form_fields(self.request, obj)

    def _get_form_class(self, obj=None):
        return (
            self.form_class or
            (
                self.core.get_rest_form_edit_class(self.request, obj)
                if obj
                else self.core.get_rest_form_add_class(self.request, obj)
            )
        )

    def _get_form_initial(self, obj):
        return {'_request': self.request, '_user': self.request.user}

    def _pre_save_obj(self, obj, form, change):
        self.core.pre_save_model(self.request, obj, form, change)

    def _save_obj(self, obj, form, change):
        self.core.save_model(self.request, obj, form, change)

    def _post_save_obj(self, obj, form, change):
        self.core.post_save_model(self.request, obj, form, change)

    def _pre_delete_obj(self, obj):
        self.core.pre_delete_model(self.request, obj)

    def _delete_obj(self, obj):
        self.core.delete_model(self.request, obj)

    def _post_delete_obj(self, obj):
        self.core.post_delete_model(self.request, obj)

    def put(self):
        # TODO: backward compatibility for bulk update should be used only patch
        return super().put() if self.kwargs.get(self.pk_name) else self.update_bulk()

    def patch(self):
        return super().patch() if self.kwargs.get(self.pk_name) else self.update_bulk()

    @transaction.smart_atomic
    def update_bulk(self):
        qs = self._filter_queryset(self._get_queryset())
        BULK_CHANGE_LIMIT = getattr(django_settings, 'BULK_CHANGE_LIMIT', 200)
        if qs.count() > BULK_CHANGE_LIMIT:
            return RestErrorResponse(
                msg=ugettext('Only %s objects can be changed by one request').format(BULK_CHANGE_LIMIT),
                code=413)

        data = self.get_dict_data()
        objects, errors = zip(*(self._update_obj(obj, data) for obj in qs))
        compact_errors = tuple(err for err in errors if err)
        return RestErrorsResponse(compact_errors) if len(compact_errors) > 0 else objects

    def _update_obj(self, obj, data):
        try:
            return (
                self._create_or_update({
                    self.pk_field_name: obj.pk,
                    **data,
                }, partial_update=True),
                None
            )
        except DataInvalidException as ex:
            return (None, self._format_message(obj, ex))
        except (ConflictException, NotAllowedException):
            raise
        except RestException as ex:
            return (None, self._format_message(obj, ex))

    def _format_message(self, obj, ex):
        return {
            'id': obj.pk,
            'errors': {k: mark_safe(force_text(v)) for k, v in ex.errors.items()} if hasattr(ex, 'errors') else {},
            '_obj_name': force_text(obj),
        }



import django
from django.conf.urls import include, url
from django.conf import settings

from is_core.site import site


urlpatterns = [url(r'^', include(site.urls))]


if settings.DEBUG:
    from django.conf.urls.static import static
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)



from django import template
from django.forms import CheckboxInput
from django.template.base import TemplateSyntaxError, token_kwargs, Node
from django.template.loader import render_to_string

from is_core.config import settings
from is_core.utils import pretty_class_name
from is_core.forms.widgets import WrapperWidget


register = template.Library()


class FormNode(Node):

    def __init__(self, template, *args, **kwargs):
        self.template = template
        self.extra_context = kwargs.pop('extra_context', {})
        super().__init__(*args, **kwargs)

    def render(self, context):
        template = self.template.resolve(context)
        # Does this quack like a Template?
        if not callable(getattr(template, 'render', None)):
            # If not, we'll try our cache, and get_template()
            template_name = template
            cache = context.render_context.dicts[0].setdefault(self, {})
            template = cache.get(template_name)
            if template is None:
                template = context.template.engine.get_template(template_name)
                cache[template_name] = template
        # Use the base.Template of a backends.django.Template.
        elif hasattr(template, 'template'):
            template = template.template
        values = {
            name: var.resolve(context)
            for name, var in self.extra_context.items()
        }
        with context.push(**values):
            return template.render(context)


@register.tag('form_renderer')
def do_form_renderer(parser, token):
    bits = token.split_contents()
    if len(bits) < 2:
        raise TemplateSyntaxError("%r tag takes at least one argument: the form that will be rendered" % bits[0])
    remaining_bits = bits[2:]
    options = token_kwargs(remaining_bits, parser, support_legacy=False)
    template_name = options.get('template', parser.compile_filter("'is_core/forms/default_form.html'"))
    options['use_csrf'] = options.get('use_csrf', parser.compile_filter('True'))
    options['form'] = parser.compile_filter(bits[1])
    options['method'] = options.get('method', parser.compile_filter("'POST'"))
    return FormNode(template_name, extra_context=options)


@register.simple_tag(takes_context=True)
def fieldset_renderer(context, form, fieldset):
    context_dict = {}
    for data in context:
        context_dict.update(data)
    request = context_dict.pop('request', None)
    values = fieldset[1]
    inline_view = values.get('inline_view_inst')
    context_dict.update({
        'class': values.get('class'),
    })

    rendered_inline_view = inline_view.render(context.new(context_dict), fieldset[0]) if inline_view else ''

    rendered_fieldsets = []
    for sub_fieldset in values.get('fieldsets', ()):
        rendered_fieldset = fieldset_renderer(context, form, sub_fieldset)
        if rendered_fieldset:
            rendered_fieldsets.append(rendered_fieldset)

    fields = values.get('fields')

    template = values.get('template') or settings.DEFAULT_FIELDSET_TEMPLATE
    context_dict.update({
        'fields': fields,
        'form': form,
        'title': fieldset[0],
        'class': values.get('class'),
        'fieldsets': values.get('fieldsets'),
        'rendered_fieldsets': rendered_fieldsets,
        'rendered_inline_view': rendered_inline_view,
    })
    return render_to_string(template, context_dict, request=request)


@register.simple_tag(takes_context=True)
def get_field(context, form, field_name):
    search_instances = []

    view = context.get('view')
    if view:
        search_instances.append(view)

    core = context.get('core')
    if core:
        search_instances.append(core)

    if hasattr(form, 'instance'):
        search_instances.append(form.instance)

    return form[field_name]


@register.simple_tag(takes_context=True)
def get_visible_fields(context, form, fieldset):
    visible_fields = []
    for field_name in fieldset:
        field = get_field(context, form, field_name)
        if not field.is_hidden:
            visible_fields.append(field)
    return visible_fields


@register.filter
def model_name(form):
    return form._meta.model._meta.module_name


@register.filter
def split(value, delimiter=','):
    return value.split(delimiter)


@register.filter
def is_checkbox(field):
    return (hasattr(field, 'field') and hasattr(field.field, 'widget') and
            (isinstance(field.field.widget, CheckboxInput) or
             (isinstance(field.field.widget, WrapperWidget) and isinstance(field.field.widget.widget, CheckboxInput))))


@register.filter
def field_type(field):
    return field.type if hasattr(field, 'type') else pretty_class_name(field.field.widget.__class__.__name__)



from is_core.generic_views.detail_views import DjangoDetailFormView
from is_core.generic_views.inlines.inline_form_views import TabularInlineFormView

from issue_tracker.models import Issue


class UserDetailView(DjangoDetailFormView):

    def leading_issue_name(self, obj):
        return obj.leading_issue.name if hasattr(obj, 'leading_issue') else 'No leading issue'
    leading_issue_name.short_description = 'leading issue name'


class SubIssuesView(TabularInlineFormView):

    model = Issue



# Django settings.
import os

try:
    from django.utils.translation import ugettext_lazy as _
except ImportError:
    def _(val): return val

PROJECT_DIR = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..', '..')
)

DEBUG = False
TEMPLATE_DEBUG = DEBUG

ADMINS = (
    # ('Your Name', 'your_email@example.com'),
)

MANAGERS = ADMINS

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# On Unix systems, a value of None will cause Django to use the same
# timezone as the operating system.
# If running in a Windows environment this must be set to the same as your
# system time zone.
TIME_ZONE = 'Europe/Prague'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'cs'

LANGUAGES = [
    ('cs', _('Czech')),
    ('en', _('English')),
    ('de', _('German')),
]

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# If you set this to False, Django will not format dates, numbers and
# calendars according to the current locale.
USE_L10N = True

# If you set this to False, Django will not use timezone-aware datetimes.
USE_TZ = True

DATE_FORMAT = 'j. E Y'

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/home/media/media.lawrence.com/media/"
MEDIA_ROOT = os.path.join(PROJECT_DIR, 'media')

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash.
# Examples: "http://media.lawrence.com/media/", "http://example.com/media/"
MEDIA_URL = '/media/'

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/home/media/media.lawrence.com/static/"
STATICFILES_ROOT = os.path.join(PROJECT_DIR, 'static')
STATIC_ROOT = STATICFILES_ROOT

# URL prefix for static files.
# Example: "http://media.lawrence.com/static/"
STATIC_URL = '/static/'

# Additional locations of static files
STATICFILES_DIRS = (

)

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)


# Make this unique, and don't share it with anybody.
SECRET_KEY = 'xe8vyy&0cw*&za++fq(%w6cx=)k53*m-@$1&pst=*oe(b#zgo+'

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
)

MIDDLEWARE = MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'is_core.middleware.RequestKwargsMiddleware',
    'is_core.middleware.HttpExceptionsMiddleware',
)

TEMPLATE_CONTEXT_PROCESSORS = (
    'django.core.context_processors.i18n',
    'django.core.context_processors.request',
    'django.core.context_processors.media',
    'django.core.context_processors.static',
    'django.core.context_processors.request',
    'django.core.context_processors.debug',
    'django.contrib.messages.context_processors.messages',
)

INTERNAL_IPS = ('127.0.0.1',)

ROOT_URLCONF = 'dj.urls'

# Python dotted path to the WSGI application used by Django's runserver.
WSGI_APPLICATION = 'dj.wsgi.application'

TEMPLATE_DIRS = (
    os.path.join(PROJECT_DIR, 'templates'),
)

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.humanize',

    # apps
    'issue_tracker',

    # IS
    'is_core',
    'block_snippets',
)

# A sample logging configuration. The only tangible logging
# performed by this configuration is to send an email to
# the site admins on every HTTP 500 error when DEBUG=False.
# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
        }
    },
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        }
    },
    'handlers': {
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler'
        },
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'standard'
        },
    },
    'loggers': {
        'django.request': {
            'handlers': ['mail_admins'],
            'level': 'ERROR',
            'propagate': True,
        },
        'is-core': {
            'handlers': ['console'],
            'level': 'DEBUG',
        }
    }
}

THROTTLING_FAILURE_VIEW = 'is_core.views.throttling.throttling_failure_view'
CSRF_FAILURE_VIEW = 'is_core.views.csrf.csrf_failure'

IS_CORE_AUTH_USE_TOKENS = True

PYSTON_CORS = True

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            # insert your TEMPLATE_DIRS here
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                # Insert your TEMPLATE_CONTEXT_PROCESSORS here or use this
                # list if you haven't customized them:
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.request',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

PYSTON_PARTIAL_PUT_UPDATE = True
PYSTON_PARTIAL_RELATED_UPDATE = True

LOGIN_URL = '/login/'

DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'



from .http_exceptions import *
from .rest_permissions import *
from .ui_ordering import *
from .ui_permissions import *
from .permissions import *
from .rest import *
from .field_permissions import *
from .utils import *
from .elasticsearch import *
from .dynamo import *



from germanium.decorators import login

from germanium.test_cases.rest import RestTestCase
from germanium.tools.trivials import assert_in, assert_equal, assert_true, assert_is_not_none
from germanium.tools.http import (assert_http_bad_request, assert_http_not_found, assert_http_method_not_allowed,
                                  assert_http_accepted, build_url, assert_http_unauthorized, assert_http_ok,
                                  assert_http_redirect)
from germanium.tools.rest import assert_valid_JSON_created_response, assert_valid_JSON_response

from .test_case import HelperTestCase, AsSuperuserTestCase

from issue_tracker.elasticsearch.models import Comment


class RESTElasticsearchTestCase(AsSuperuserTestCase, HelperTestCase, RestTestCase):

    COMMENT_API_URL = '/api/elasticsearch-comment/'
    COMMENT_UI_URL = '/elasticsearch-comment/'

    @classmethod
    def setUpClass(cls):
        for i in range(10):
            comment = Comment(
                user_id=str(i),
                content=f'test message {i}',
                is_public=bool(i % 2),
                priority=i
            )
            comment.meta.id = i
            comment.save()
        Comment._index.refresh()

    @classmethod
    def tearDownClass(cls):
        Comment._index.delete()

    @login(is_superuser=True)
    def test_get_api_elasticsearch_comments_should_return_right_data(self):
        resp = self.get(f'{self.COMMENT_API_URL}?_fields=id,user_id,content,is_public,priority')
        assert_valid_JSON_response(resp)
        assert_equal(len(resp.json()), 10)
        for i, data in enumerate(resp.json()):
            assert_equal(data['priority'], i)
            assert_equal(data['user_id'], str(i))
            assert_equal(data['is_public'], bool(i % 2))
            assert_equal(data['content'], f'test message {i}')
            assert_equal(data['id'], str(i))

    @login(is_superuser=True)
    def test_get_api_elasticsearch_one_comment_should_return_right_data(self):
        for i in range(10):
            resp = self.get(f'{self.COMMENT_API_URL}{i}/?_fields=id,user_id,content,is_public,priority')
            assert_valid_JSON_response(resp)
            data = resp.json()
            assert_equal(data['priority'], i)
            assert_equal(data['user_id'], str(i))
            assert_equal(data['is_public'], bool(i % 2))
            assert_equal(data['content'], f'test message {i}')
            assert_is_not_none(data['id'])

    def test_get_api_elasticsearch_comments_without_authorization_should_return_unauthorized(self):
        assert_http_unauthorized(self.get(self.COMMENT_API_URL))

    @login(is_superuser=True)
    def test_get_elasticsearch_should_return_list_view(self):
        assert_http_ok(self.get(self.COMMENT_UI_URL))

    @login(is_superuser=True)
    def test_get_elasticsearch_should_return_detail_view(self):
        for i in range(10):
            assert_http_ok(self.get(f'{self.COMMENT_UI_URL}{i}/'))

    @login(is_superuser=True)
    def test_get_elasticsearch_should_return_not_found_for_invalid_object_detail(self):
        assert_http_not_found(self.get(f'{self.COMMENT_UI_URL}10/'))

    def test_get_elasticsearch_comments_without_authorization_should_return_redirect_to_login(self):
        assert_http_redirect(self.get(f'{self.COMMENT_UI_URL}0/'))
        assert_http_redirect(self.get(self.COMMENT_UI_URL))



from django.conf import settings as django_settings


def _get_auth_login_view():
    if ('auth_token' in django_settings.INSTALLED_APPS
            and getattr(django_settings, 'AUTH_TOKEN_TWO_FACTOR_ENABLED', False)):
        return 'auth_token.contrib.is_core_auth.views.TwoFactorLoginView'
    elif 'auth_token' in django_settings.INSTALLED_APPS:
        return 'auth_token.contrib.is_core_auth.views.LoginView'
    else:
        return 'is_core.views.auth.LoginView'


DEFAULTS = {
    'AUTH_LOGIN_CODE_VERIFICATION_VIEW': (
        'auth_token.contrib.is_core_auth.views.LoginCodeVerificationView'
        if ('auth_token' in django_settings.INSTALLED_APPS
            and getattr(django_settings, 'AUTH_TOKEN_TWO_FACTOR_ENABLED', False))
        else None
    ),
    'AUTH_FORM_CLASS': (
        'auth_token.contrib.is_core_auth.forms.TokenAuthenticationSmartForm'
        if 'auth_token' in django_settings.INSTALLED_APPS
        else 'django.contrib.auth.forms.AuthenticationForm'
    ),
    'AUTH_RESOURCE_CLASS': None,
    'AUTH_LOGIN_VIEW': _get_auth_login_view(),
    'AUTH_LOGOUT_VIEW': (
        'auth_token.contrib.is_core_auth.views.LogoutView' if 'auth_token' in django_settings.INSTALLED_APPS
        else 'is_core.views.auth.LogoutView'
    ),
    'CODE_VERIFICATION_URL': '/login/login-code-verification/',
    'HOME_CORE': 'is_core.main.HomeUiCore',
    'HOME_VIEW': 'is_core.generic_views.base.HomeView',
    'MENU_GENERATOR': 'is_core.menu.MenuGenerator',
    'USERNAME': 'username',
    'PASSWORD': 'password',
    'LOGIN_URL': '/login/',
    'LOGOUT_URL': '/logout/',
    'LOGIN_API_URL': lambda s: '/api{}'.format(s.LOGIN_URL),
    'EXPORT_TYPES': '',
    'FOREIGN_KEY_MAX_SELECTBOX_ENTRIES': 500,
    'LIST_PER_PAGE': 20,
    'REST_PAGINATOR_MAX_TOTAL': 10000,
    'RESPONSE_EXCEPTION_FACTORY': 'is_core.exceptions.response.ui_rest_response_exception_factory',
    'DEFAULT_FIELDSET_TEMPLATE': 'is_core/forms/default_fieldset.html',
    'HEADER_IMAGE': None,
    'ENVIRONMENT': '',
    'BACKGROUND_EXPORT_TASK_TIME_LIMIT': 60*60,
    'BACKGROUND_EXPORT_TASK_SOFT_TIME_LIMIT': (60*60) - 5,
    'BACKGROUND_EXPORT_TASK_UPDATE_REQUEST_FUNCTION': None,
    'BACKGROUND_EXPORT_TASK_CALLBACK': None,
    'BACKGROUND_EXPORT_TASK_CALLBACK_HANDLED_EXCEPTIONS': (),
    'BACKGROUND_EXPORT_TASK_QUEUE': None,
    'BACKGROUND_EXPORT_SERIALIZATION_LIMIT': 2000,
    'BACKGROUND_EXPORT_STORAGE_CLASS': 'django.core.files.storage.DefaultStorage',
    'BACKGROUND_EXPORT_EXPIRATION_DAYS': 30,
    'COLUMN_MANAGER': True,
}


class Settings:

    def __getattr__(self, attr):
        if attr not in DEFAULTS:
            raise AttributeError('Invalid IS_CORE setting: "{}"'.format(attr))

        default = DEFAULTS[attr]
        return getattr(django_settings, 'IS_CORE_{}'.format(attr), default(self) if callable(default) else default)


settings = Settings()
